'''
Various functions used by SKHASH that do not fall into one of the other file categories.
'''

# External libraries
import numpy as np
import pandas as pd


def read_control_file(control_filepath,p_dict):
    '''
    Reads the user's control file.
    If the control file does not contain a variable, the default value in p_dict is used.
    If the control file contains an unexpected variable, it is ignored after printing a warning message.
    This file should follow the format:
        $var_name
        var_value
        ...
    Where "$var_name" is the name of the variable with a "$" as the first non-whitespace character,
    and "var_value" is the value of the variable on the following line.
    Velocity models ($vmodel_paths) can be specified on multiple lines if desired, e.g.:
        $vmodel_paths
        path_to/vmodel1.txt
        path_to/vmodel2.txt

    User text on all other lines will be ignored, provided the first non-whitespace character on the line is not a "$".
    User text will also be ignored following a space (" ") after "$var_name"
    '''
    # Reads lines from the control file
    with open(control_filepath) as f:
        lines=f.read().splitlines()
        lines=[line.strip() for line in lines]

    # Removes any text following a '#'
    for line_x in range(len(lines)):
        if '#' in lines[line_x]:
            lines[line_x]=lines[line_x][:lines[line_x].find('#')].strip()

    # Identifies the variable names and their line numbers
    var_name=[]
    var_line=[]
    value_line=[]
    for line_x,line in enumerate(lines):
        if line: # If not empty
            if line[0]=='$':
                tmp_var_name=line.split()[0][1:]
                tmp_var_name=tmp_var_name.split('#')[0] # In case the user didn't put a space after the variable name.
                if tmp_var_name in p_dict.keys():
                    var_name.append(tmp_var_name)
                    var_line.append(line_x)
                else:
                    raise ValueError('Unknown variable name ({}) in the control file on line {}:\n\t{}'.format(tmp_var_name,line_x,line))
            elif line[0]!='#':
                value_line.append(line_x)

    # Ensures there are not duplicate variables
    var_set=set()
    duplicate_var=[x for x in var_name if x in var_set or var_set.add(x)]
    if duplicate_var:
        raise ValueError('Duplicate variable name in control file: \"{}\"'.format(duplicate_var[0]))

    # Handles velocity model paths that are entered on multiple lines
    if 'vmodel_paths' in var_name:
        tmp_ind=var_name.index('vmodel_paths')
        var_tmp_ind=[x for x in value_line if (x > var_line[tmp_ind])]
        if len(var_tmp_ind)>1:
            if (tmp_ind+1)<len(var_line):
                var_tmp_ind=[x for x in var_tmp_ind if (x < var_line[tmp_ind+1])]
            vmodel_paths_join=' '.join([lines[x] for x in var_tmp_ind])

            value_line=[x for x in value_line if x not in var_tmp_ind[1:]]
            lines[var_tmp_ind[0]]=vmodel_paths_join

    # Ensures that every value line corresponds with a variable
    corr_bool=np.isin( (np.asarray(value_line)-1),var_line)
    if not(np.all(corr_bool)):
        prob_ind=np.where(corr_bool==False)[0][0]
        raise ValueError('Unable to associate a value (\"{}\") on line {} with a variable in the control file. If this is a variable, start the line with a \'$\'. If this is intended to be ignored, start the line with a \'#\'.'.format(lines[value_line[prob_ind]],value_line[prob_ind]))

    # Determines the variable values that follow the variable name lines
    var_value=[]
    for line_x in var_line:
        tmp_var_value=lines[line_x+1].strip()

        if tmp_var_value[0]=='#':
            raise ValueError('Value (\"{}\") on line {} in the control file starts with a \'#\'.'.format(tmp_var_value,line_x+1) )
        if tmp_var_value.isdigit(): # if int
            tmp_var_value=int(tmp_var_value)
        elif tmp_var_value.replace('.','',1).lstrip('-').isdigit(): # if float
            tmp_var_value=float(tmp_var_value)
        elif (tmp_var_value=='True'):
            tmp_var_value=True
        elif (tmp_var_value=='False'): # if bool
            tmp_var_value=False
        elif (' ' in tmp_var_value): # if list
            tmp_var_value=tmp_var_value.split()
        var_value.append(tmp_var_value)

    # Assigns the variable values to the dictionary
    for name,value in zip(var_name,var_value):
        p_dict[name]=value

    # Some basic variable type fixing
    p_dict['look_dep']=[int(x) for x in p_dict['look_dep']]
    p_dict['look_del']=[int(x) for x in p_dict['look_del']]
    if type(p_dict['vmodel_paths'])==str:
        p_dict['vmodel_paths']=[p_dict['vmodel_paths']]

    return p_dict


def read_simulps(simulpsfile,pol_df):
    '''
    Appends the takeoffs & azimuths from a SIMULPS 3D (Evans et al., 1994) file
    to the polarity dataframe.
    '''
    duplicated_sta_eventid_df=pol_df[pol_df.duplicated(subset=['station','event_id'])]
    if not(duplicated_sta_eventid_df.empty):
        raise ValueError('When using a SIMULPS file (simulpsfile: {}), the station codes for each event in the polarity files must be unique. Example polarity problems:\n{}'.format(simulpsfile,duplicated_sta_eventid_df.head(10)))

    simulps_df=read_simulps_file(simulpsfile)
    # pol_df=pol_df.merge(simulps_df,on=['station','event_id'])

    duplicated_simulps_df=simulps_df[simulps_df.duplicated(subset=['station','event_id'])]
    if not(duplicated_simulps_df.empty):
        print('*WARNING: The SIMULPS file (simulpsfile: {}) contains duplicate station codes for an individual event. Only the first station record for each event will be considered.'.format(simulpsfile))
        simulps_df=simulps_df.drop(duplicated_simulps_df.index).reset_index(drop=True)

    # Updates the source-receiver distances, takeoff, and azimuths. Information in SIMULPS file will overwrite any previous distances/angles.
    pol_df.update(pol_df.drop(columns=['sr_dist_km', 'azimuth', 'takeoff', 'azimuth_uncertainty', 'takeoff_uncertainty']).merge(simulps_df, how='left', on=['station','event_id']))

    return pol_df


def read_simulps_file(simulpsfile):
    '''
    Reads a SIMULPS 3D (Evans et al., 1994) file. See HASH v1.2 manual for the required formatting.
    '''
    col_ind_name=np.asarray([
        [1,3,'year'],
        [3,5,'month'],
        [5,7,'day'],
        [8,10,'hour'],
        [10,12,'minute'],
        [13,18,'second'],
        [19,21,'lat_deg'],
        [21,22,'latNS'],
        [22,27,'lat_min'],
        [28,31,'lon_deg'],
        [31,31,'lonEW'],
        [32,37,'lon_min'],
        [38,44,'depth'],
        [47,50,'mag_pref'],
        [55,63,'event_id'],
    ])
    col_ind=col_ind_name[:,0:2].astype(int)

    pick_col_ind_name=np.asarray([
        [1,5,'station'],
        [6,11,'sr_dist_km'],
        [11,15,'azimuth'],
        [15,19,'takeoff'],
        [65,69,'azimuth_uncertainty'],
        [69,72,'takeoff_uncertainty'],
    ])
    pick_col_ind=pick_col_ind_name[:,0:2].astype(int)
    pick_col_name=pick_col_ind_name[:,2]

    with open(simulpsfile, 'r') as file1:
        lines = file1.read().splitlines()
    header_line_nums=[]
    station_startline_nums=[]
    station_endline_nums=[]
    station_line_flag=False
    for line_x in range(len(lines)):
        if lines[line_x][:16]=='  DATE    ORIGIN':
            header_line_nums.append(line_x)
        elif lines[line_x][:11]=='  STN  DIST':
            station_startline_nums.append(line_x+1)
            station_line_flag=True
        elif station_line_flag & (lines[line_x].strip()==''):
            station_endline_nums.append(line_x-1)
            station_line_flag=False
    if station_line_flag:
        station_endline_nums.append(line_x-1)

    if len(station_startline_nums)!=len(station_endline_nums):
        raise ValueError('Error reading simulpsfile ({}).'.format(simulpsfile))

    header_all=[]
    event_id=[]
    for header_x in range(len(header_line_nums)):
        line=lines[header_line_nums[header_x]+1]
        header_all.append([line[i:j] for i,j in zip(col_ind[:,0], col_ind[:,1])])
        event_id.append(header_all[-1][-1].strip())

    simulps_all=[]
    for event_x in range(len(station_startline_nums)):
        tmp_event=[]
        for line_x in range(station_startline_nums[event_x],station_endline_nums[event_x]+1):
            tmp_event.append([lines[line_x][i:j] for i,j in zip(pick_col_ind[:,0], pick_col_ind[:,1])])
        tmp_pick_df=pd.DataFrame(tmp_event,columns=pick_col_name)
        tmp_pick_df['event_id']=event_id[event_x]
        simulps_all.append(tmp_pick_df)
    simulps_df=pd.concat(simulps_all).reset_index(drop=True)

    simulps_df=simulps_df.astype({'sr_dist_km':float,'azimuth':int,'takeoff':int,'azimuth_uncertainty':float,'takeoff_uncertainty':float,'event_id':str})
    simulps_df['station']=simulps_df['station'].str.strip()
    simulps_df['takeoff']=180-simulps_df['takeoff']
    return simulps_df


def read_catalog_file(p_dict):
    '''
    Reads catalog file
    '''
    cat_df=pd.read_csv(p_dict['catfile'],skipinitialspace=True,comment='#')
    if not({'time','event_id','latitude','longitude','depth'}.issubset(cat_df.columns)):
        raise ValueError(('The catalog file (catfile: {}) must contain the following column names: [time, event_id, latitude, longitude, depth]\n'+
                        'Only the following columns were provided:\n{}').format(p_dict['catfile'],cat_df.columns.values))
    cat_df=cat_df.rename(columns={'time':'origin_DateTime','latitude':'origin_lat','longitude':'origin_lon','depth':'origin_depth_km','mag':'event_mag'})

    if not('vert_uncert_km' in cat_df.columns):
        cat_df['vert_uncert_km']=p_dict['default_vert_uncert_km']
    if not('horz_uncert_km' in cat_df.columns):
        cat_df['horz_uncert_km']=p_dict['default_horz_uncert_km']

    if 'origin_DateTime' in cat_df.columns:
        cat_df['origin_DateTime']=pd.to_datetime(cat_df['origin_DateTime'])
    cat_df=cat_df.filter(['origin_DateTime','origin_lat','origin_lon','origin_depth_km','horz_uncert_km','vert_uncert_km','event_mag','event_id'])

    # Ensures column data formats are the expected type
    cat_df['origin_lat']=cat_df['origin_lat'].astype(float)
    cat_df['origin_lon']=cat_df['origin_lon'].astype(float)
    cat_df['origin_depth_km']=cat_df['origin_depth_km'].astype(float)
    cat_df['horz_uncert_km']=cat_df['horz_uncert_km'].astype(float)
    cat_df['vert_uncert_km']=cat_df['vert_uncert_km'].astype(float)
    cat_df['event_id']=cat_df['event_id'].astype(str)

    cat_df=cat_df.sort_values(by='event_id').reset_index(drop=True)

    return cat_df



'''
Functions used to provide quality control of user inputs
'''

# Standard libraries
import os
import multiprocessing

# External libraries
import numpy as np
import pandas as pd


def check_input_params(p_dict,qual_criteria_dict):
	'''
	Performs various checks to ensure that the user-defined parameters are appropriate
	'''
	# Ensures input format is lowercase and is not empty
	if (p_dict['input_format']==''):
		p_dict['input_format']='skhash'
	else:
		p_dict['input_format']=p_dict['input_format'].lower()

	for format_file in ['input_format_stfile','input_format_fpfile','input_format_impfile','input_format_conpfile','input_format_dlpfile']:
		p_dict[format_file]=p_dict[format_file].lower()

	if (p_dict['input_format'][:-1]=='hash') & (p_dict['catfile']!=''):
		raise ValueError('When using the {} input format, no catalog file (catfile) should be provided.'.format(p_dict['input_format']))

	if (p_dict['input_format'][:-1]=='hash') & (p_dict['fpfile']==''):
		raise ValueError('When using the {} input format, a p-polarity file (fpfile) should be provided.'.format(p_dict['input_format']))

	if (p_dict['input_format']=='hash1') | (p_dict['input_format']=='hash5'):
		if len(p_dict['vmodel_paths'])>0:
			raise ValueError('When using the {} input format, no velocity model(s) (vmodel_paths) should be provided.'.format(p_dict['input_format']))

	if (p_dict['input_format']=='hash1'): # hash1 only matches on the station and channel codes
		p_dict['require_network_match']=False
		p_dict['require_station_match']=True
		p_dict['require_location_match']=False
		p_dict['require_channel_match']=True
	elif p_dict['input_format'] in ['hash2','hash3','hash4','hash5']: # hash2-5 only matches on the network, station, and channel codes
		p_dict['require_network_match']=True
		p_dict['require_station_match']=True
		p_dict['require_location_match']=False
		p_dict['require_channel_match']=True

	default_input_format=p_dict['input_format']

	# Ensures the stfile format is an accepted format
	if p_dict['stfile']:
		possible_format=['skhash','hypoinverse','hash1','hash2','hash3','hash4','hash5']
		if p_dict['input_format_stfile']=='':
			p_dict['input_format_stfile']=default_input_format
		if not(p_dict['input_format_stfile'] in possible_format):
			raise ValueError('input_format_stfile must be one of the following:\n\t{}'.format(possible_format))

	# Ensures the fpfile format is an accepted format
	if p_dict['fpfile']:
		possible_format=['skhash','ncsn','hypoinverse','hash1','hash2','hash3','hash4','hash5','quakeml']
		if p_dict['input_format_fpfile']=='':
			p_dict['input_format_fpfile']=default_input_format
		if not(p_dict['input_format_fpfile'] in possible_format):
			raise ValueError('input_format_fpfile must be one of the following:\n\t{}'.format(possible_format))

	# Ensures the impfile format is an accepted format
	if p_dict['impfile']:
		possible_format=['skhash','hash1','hash2','hash3','hash4','hash5']
		if p_dict['input_format_impfile']=='':
			p_dict['input_format_impfile']=default_input_format
		if not(p_dict['input_format_impfile'] in possible_format):
			raise ValueError('input_format_impfile must be one of the following:\n\t{}'.format(possible_format))

	# Ensures the conpfile format is an accepted format
	if p_dict['conpfile']:
		possible_format=['skhash','hash1','hash2','hash3','hash4','hash5']
		if p_dict['input_format_conpfile']=='':
			p_dict['input_format_conpfile']=default_input_format
		else:
			p_dict['input_format_conpfile']=p_dict['input_format_conpfile'].lower()
		if not(p_dict['input_format_conpfile'] in possible_format):
			raise ValueError('input_format_conpfile must be one of the following:\n\t{}'.format(possible_format))

	# Ensures the dlpfile format is an accepted format
	if p_dict['dlpfile']:
		possible_format=['skhash','hash1','hash2','hash3','hash4','hash5']
		if p_dict['input_format_dlpfile']=='':
			p_dict['input_format_dlpfile']=default_input_format
		else:
			p_dict['input_format_dlpfile']=p_dict['input_format_dlpfile'].lower()
		if not(p_dict['input_format_dlpfile'] in possible_format):
			raise ValueError('input_format_dlpfile must be one of the following:\n\t{}'.format(possible_format))

	# Ensures the plfile format is an accepted format
	if p_dict['plfile']:
		possible_format=['skhash','hash1','hash2','hash3','hash4','hash5']
		if p_dict['input_format_plfile']=='':
			p_dict['input_format_plfile']=default_input_format
		else:
			p_dict['input_format_plfile']=p_dict['input_format_plfile'].lower()
		if not(p_dict['input_format_plfile'] in possible_format):
			raise ValueError('input_format_plfile must be one of the following:\n\t{}'.format(possible_format))

	# Ensures the ampfile format is an accepted format
	if p_dict['ampfile']:
		possible_format=['skhash','hash3']
		if (p_dict['input_format_ampfile']==''):
			p_dict['input_format_ampfile']=default_input_format
		else:
			p_dict['input_format_ampfile']=p_dict['input_format_ampfile'].lower()
		if not(p_dict['input_format_ampfile'] in possible_format):
			raise ValueError('input_format_ampfile must be one of the following:\n\t{}'.format(possible_format))

	# Ensures the relative ampfile format is an accepted format
	if p_dict['relampfile']:
		possible_format=['skhash','hash3']
		if (p_dict['input_format_relampfile']==''):
			p_dict['input_format_relampfile']=default_input_format
		else:
			p_dict['input_format_relampfile']=p_dict['input_format_relampfile'].lower()
		if not(p_dict['input_format_relampfile'] in possible_format):
			raise ValueError('input_format_relampfile must be one of the following:\n\t{}'.format(possible_format))

	# Determines what station information should be used to associate picks with metadata.
	p_dict['merge_on']=[]
	if p_dict['require_network_match']:
		p_dict['merge_on'].append('network')
	if p_dict['require_station_match']:
		p_dict['merge_on'].append('station')
	if p_dict['require_location_match']:
		p_dict['merge_on'].append('location')
	if p_dict['require_channel_match']:
		p_dict['merge_on'].append('channel')

	# Ensures that require_location_match=False if doing HASH format as they do not consider location codes
	if (p_dict['input_format'][:-1]=='hash') & (p_dict['require_location_match']):
		print('*WARNING: Metadata cannot be associated using location codes (require_location_match=True) when using the {} input_format. Setting require_location_match=False.'.format(p_dict['input_format']))
		p_dict['require_location_match']=False

	if p_dict['input_format']=='skhash':
		if not(p_dict['require_network_match'] | p_dict['require_station_match'] | p_dict['require_location_match'] | p_dict['require_channel_match']):
			raise ValueError('When using the SKHASH input format, at least one of the following metadata matches must be True: require_network_match, require_station_match, require_location_match, require_channel_match')

	if (not(p_dict['fpfile'])) & (not(p_dict['impfile'])) & (not(p_dict['conpfile'])) & (not(p_dict['dlpfile'])):
		print('*WARNING: No P polarity files provided (fpfile, impfile, conpfile, dlpfile), so there will be no polarity agreement output file (outfile_pol_agree: {}).'.format(p_dict['outfile_pol_agree']))
		p_dict['outfile_pol_agree']=''
	if (not(p_dict['ampfile'])) & (not(p_dict['relampfile'])) & (p_dict['corfile']!=''):
		print('*WARNING: No traditional or consensus S/P amplitude file provided (ampfile or relampfile), so the S/P correction file (corfile: {}) will be ignored.'.format(p_dict['corfile']))
		p_dict['corfile']=''

	# For any filepath vars, replaces '~' with user's home directory
	for path_var in ['controlfile','catfile','stfile','plfile','corfile','fpfile','impfile','conpfile',
				  	'dlpfile','ampfile','relampfile','simulpsfile','outfile1','outfile2','outfile_pol_agree',
					'outfile_sp_agree','outfile_pol_info','outfolder_plots']:
		if p_dict[path_var]:
			if p_dict[path_var][0]=='~':
				p_dict[path_var]=os.path.expanduser(p_dict[path_var])
	if p_dict['vmodel_paths']:
		for model_x in range(len(p_dict['vmodel_paths'])):
			if p_dict['vmodel_paths'][model_x][0]=='~':
				p_dict['vmodel_paths'][model_x]=os.path.expanduser(p_dict['vmodel_paths'][model_x])

	# Ensures all provided input files exist
	if p_dict['catfile']:
		if not(os.path.isfile(p_dict['catfile'])):
			raise ValueError('Catalog file (catfile) does not exist:{}'.format(p_dict['catfile']))
	if p_dict['stfile']:
		if not(os.path.isfile(p_dict['stfile'])):
			raise ValueError('Station file (stfile) does not exist:{}'.format(p_dict['stfile']))
	if p_dict['plfile']:
		if not(os.path.isfile(p_dict['plfile'])):
			raise ValueError('Polarity reversal file (plfile) does not exist:{}'.format(p_dict['plfile']))
	if p_dict['corfile']:
		if not(os.path.isfile(p_dict['corfile'])):
			raise ValueError('Station correction file (corfile) does not exist:{}'.format(p_dict['corfile']))
	if p_dict['ampfile']:
		if not(os.path.isfile(p_dict['ampfile'])):
			raise ValueError('S/P amplitude file (ampfile) does not exist:{}'.format(p_dict['ampfile']))
	if p_dict['relampfile']:
		if not(os.path.isfile(p_dict['relampfile'])):
			raise ValueError('Relative S/P amplitude file (relampfile) does not exist:{}'.format(p_dict['relampfile']))
	if p_dict['fpfile']:
		if not(os.path.isfile(p_dict['fpfile'])):
			raise ValueError('P-polarity file (fpfile) does not exist:{}'.format(p_dict['fpfile']))
	if p_dict['impfile']:
		if not(os.path.isfile(p_dict['impfile'])):
			raise ValueError('Imputed P-polarity file (impfile) does not exist:{}'.format(p_dict['impfile']))
	if p_dict['conpfile']:
		if not(os.path.isfile(p_dict['conpfile'])):
			raise ValueError('Consensus P-polarity file (conpfile) does not exist:{}'.format(p_dict['conpfile']))
	# Ensures the velocity model paths also exist:
	for tmp_vmodel_path in p_dict['vmodel_paths']:
		if not(os.path.isfile(tmp_vmodel_path)):
			raise ValueError('Velocity model (vmodel_paths) path does not exist: {}'.format(tmp_vmodel_path))

	# Ensures the output folder for plots exists
	if p_dict['outfolder_plots']:
		if os.path.exists(p_dict['outfolder_plots']):
			if not(os.path.isdir(p_dict['outfolder_plots'])):
				raise ValueError('Folder containg output plots (outfolder_plots) exists but is not a folder: {}'.format(p_dict['outfolder_plots']))

	# If plotting station names, ensures station names are being considered
	if p_dict['outfolder_plots']:
		if (p_dict['plot_station_names'] & (p_dict['require_station_match']==False)):
			raise ValueError('Station names can be added to plots ($plot_station_names==True) only if you are considering station names ($require_station_match==True)')

	# Ensures all of the input and output filepaths are unique
	filepath_vars=[
		'catfile',
		'stfile',
		'plfile',
		'corfile',
		'fpfile',
		'impfile',
		'conpfile',
		'ampfile',
		'relampfile',
		'simulpsfile',
		'outfile1',
		'outfile2',
		'outfile_pol_agree',
		'outfile_sp_agree',
		'outfile_pol_info',
		'outfolder_plots']
	tmp_dict = {key: p_dict[key] for key in filepath_vars if p_dict[key]!=''}
	if len(tmp_dict)!=len(set(tmp_dict.values())):
		rev_multidict = {}
		for key, value in tmp_dict.items():
			rev_multidict.setdefault(value, set()).add(key)
		tmp=[values for key, values in rev_multidict.items() if len(values) > 1]
		raise ValueError('Filepaths for the following variables are repeated: {}'.format(tmp[0]))

	# Ensures existing files are not overwritten by the output, if desired
	if (not(p_dict['overwrite_output_file'])):
		if p_dict['outfile1']:
			if os.path.exists(p_dict['outfile1']):
				raise ValueError('Preferred mechanism output file (outfile1={}) already exists. Either change the outfile1 path, remove this file, or set overwrite_output_file=True.'.format(p_dict['outfile1']))
		if p_dict['outfile2']:
			if os.path.exists(p_dict['outfile2']):
				raise ValueError('Acceptable mechanism output file (outfile2={}) already exists. Either change the outfile1 path, remove this file, or set overwrite_output_file=True.'.format(p_dict['outfile2']))
		if p_dict['outfile_pol_agree']:
			if os.path.exists(p_dict['outfile_pol_agree']):
				raise ValueError('Polarity agreement output file (outfile_pol_agree={}) already exists. Either change the outfile1 path, remove this file, or set overwrite_output_file=True.'.format(p_dict['outfile_pol_agree']))
		if p_dict['outfile_sp_agree']:
			if os.path.exists(p_dict['outfile_sp_agree']):
				raise ValueError('S/P difference output file (outfile_sp_agree={}) already exists. Either change the outfile1 path, remove this file, or set overwrite_output_file=True.'.format(p_dict['outfile_sp_agree']))
		if p_dict['outfile_pol_info']:
			if os.path.exists(p_dict['outfile_pol_info']):
				raise ValueError('Polarity info output file (outfile_pol_info={}) already exists. Either change the outfile1 path, remove this file, or set overwrite_output_file=True.'.format(p_dict['outfile_pol_info']))

	# Prints a warning if a quakeml file is being read in which the takeoffs are not being flipped
	if p_dict['fpfile']:
		if (p_dict['input_format_fpfile']=='quakeml') and (p_dict['flip_takeoff']==False):
			print('*WARNING: The polarity file ({}) is in the QuakeML format and $flip_takeoff is False. The QuakeML standard indicates takeoff values of < 90 are downgoing, the opposite notation of SKHASH. Should $flip_takeoff be True?'.format(p_dict['fpfile']))

	# Creates output directories if necessary
	if p_dict['outfile1']:
		folder_path=os.path.dirname(p_dict['outfile1'])
		if folder_path:
			os.makedirs(folder_path,exist_ok=True)
	if p_dict['outfile2']:
		folder_path=os.path.dirname(p_dict['outfile2'])
		if folder_path:
			os.makedirs(folder_path,exist_ok=True)
	if p_dict['outfile_pol_agree']:
		folder_path=os.path.dirname(p_dict['outfile_pol_agree'])
		if folder_path:
			os.makedirs(folder_path,exist_ok=True)
	if p_dict['outfile_sp_agree']:
		folder_path=os.path.dirname(p_dict['outfile_sp_agree'])
		if folder_path:
			os.makedirs(folder_path,exist_ok=True)
	if p_dict['outfile_pol_info']:
		folder_path=os.path.dirname(p_dict['outfile_pol_info'])
		if folder_path:
			os.makedirs(folder_path,exist_ok=True)
	if p_dict['outfolder_plots']:
		os.makedirs(p_dict['outfolder_plots'],exist_ok=True)

	if p_dict['min_amp']>1:
		raise ValueError('min_amp must be a value <=1')
	if (p_dict['max_agap']<0) | (p_dict['max_agap']>360):
		raise ValueError('The max azimuthal gap (max_agap) must be a value between 0 and 360')
	if (p_dict['max_pgap']<0) | (p_dict['max_pgap']>90):
		raise ValueError('The max takeoff angle gap (max_pgap) must be a value between 0 and 90')
	if p_dict['maxout']<1:
		raise ValueError('The max number of acceptable focal mechanisms (maxout) must be at least 1 (ideally larger!).')
	if p_dict['nmc']<1:
		raise ValueError('The number of trials (nmc) must be at least 1 (ideally larger!).')

	if p_dict['min_quality_report']:
		if not(p_dict['min_quality_report'] in qual_criteria_dict['qual_letter']):
			raise ValueError('The minimum mech quality (min_quality_report: {}) must be one of the quality codes:\n\t{}'.format(p_dict['min_quality_report'],qual_criteria_dict['qual_letter']))

	if p_dict['delmax']>p_dict['look_del'][1]:
		raise ValueError('The maximum source-receiver distance (delmax: {} km) is greater than the model distance range (look_del: {}-{} km)'.format(p_dict['delmax'],p_dict['look_del'][0],p_dict['look_del'][1]))

	# Makes sure the number of requested cores are available
	if p_dict['num_cpus']!=1:
		num_cpus_avail=multiprocessing.cpu_count()-1
		if (p_dict['num_cpus']==0): # Uses all available cores
			p_dict['num_cpus']=num_cpus_avail
		elif (p_dict['num_cpus']>num_cpus_avail): # If the num requested > available cores, use all available
			print('You requested {} cores, but only {} are available. Using all available cores instead.'.format(p_dict['num_cpus'],num_cpus_avail))
			p_dict['num_cpus']=num_cpus_avail

	if p_dict['outfolder_plots']:
		try:
			import matplotlib.pyplot as plt
		except:
			raise ImportError('matplotlib was unable to be loaded, but it is required to make plots. Either install or remove $outfolder_plots from your control file.')

	return p_dict


def check_cat_df(cat_df,pol_df,p_dict):
	'''
	Does quality control testing on the catalog dataframe (cat_df).
	'''
	if len(cat_df)==0:
		return pd.DataFrame()

	if not('vert_uncert_km' in cat_df.columns):
		cat_df['vert_uncert_km']=0
	if not('horz_uncert_km' in cat_df.columns):
		cat_df['horz_uncert_km']=0

	# Ensures the event_id column is string
	cat_df['event_id']=cat_df['event_id'].astype(str)

	# Rounds lat/lon to the specified number of decimal places
	if p_dict['epicenter_degree_precision']>0:
		cat_df['origin_lon']=cat_df['origin_lon'].round(p_dict['epicenter_degree_precision'])
		cat_df['origin_lat']=cat_df['origin_lat'].round(p_dict['epicenter_degree_precision'])

	# Ensures the vertical and horizontal uncertainties are positive
	if np.any(cat_df['vert_uncert_km']<0):
		prob_ind=cat_df['vert_uncert_km'].argmin()
		raise ValueError('Vertical uncertainties must be >= 0. Example problem: event {}, vert uncertainty: {} km'.
			format(cat_df.loc[prob_ind,'event_id'],cat_df.loc[prob_ind,'vert_uncert_km']))
	if np.any(cat_df['horz_uncert_km']<0):
		prob_ind=cat_df['horz_uncert_km'].argmin()
		raise ValueError('Horizontal uncertainties must be >= 0. Example problem: event {}, horz uncertainty: {} km'.
			format(cat_df.loc[prob_ind,'event_id'],cat_df.loc[prob_ind,'horz_uncert_km']))

	if p_dict['perturb_epicentral_location']:
		if (p_dict['default_vert_uncert_km']==0) & (cat_df['vert_uncert_km'].min()==0):
			print('*WARNING: You set default_vert_uncert_km=0 and there are events with 0 km uncertainties. The hypocentral perturbation will not work correctly.')
		else:
			novert_uncert_ind=np.where(cat_df['vert_uncert_km']==0)[0]
			if len(novert_uncert_ind)>0:
				print('{} events have a 0 km vertical uncertainty. Setting the uncertainty to {} km for these events'.
					format(len(novert_uncert_ind),p_dict['default_vert_uncert_km']))
				cat_df.loc[novert_uncert_ind,'vert_uncert_km']=p_dict['default_vert_uncert_km']
		if (p_dict['default_horz_uncert_km']==0) & (cat_df['horz_uncert_km'].min()==0):
			print('*WARNING: You set default_horz_uncert_km=0 and there are events with 0 km uncertainties. The hypocentral perturbation will not work correctly.')
		else:
			nohorz_uncert_ind=np.where(cat_df['horz_uncert_km']==0)[0]
			if len(nohorz_uncert_ind)>0:
				print('{} events have a 0 km horizontal uncertainty. Setting the uncertainty to {} km for these events'.
					format(len(nohorz_uncert_ind),p_dict['default_horz_uncert_km']))
				cat_df.loc[nohorz_uncert_ind,'horz_uncert_km']=p_dict['default_horz_uncert_km']

	# Selects only cataloged events that have a polarity measurement
	cat_df=cat_df.loc[cat_df['event_id'].isin(pol_df['event_id']),:].reset_index(drop=True)

	# Rounds magnitude
	if 'event_mag' in cat_df.columns:
		cat_df['event_mag']=cat_df['event_mag'].round(8)

	return cat_df


def check_pol_df(pol_df,cat_df,p_dict):
	'''
	Does quality control testing on the polarity dataframe (pol_df).
	'''
	# Ensures the event_id column is string
	pol_df['event_id']=pol_df['event_id'].astype(str)

	if not('p_polarity' in pol_df.columns):
		pol_df['p_polarity']=np.nan
	if not('sr_dist_km' in pol_df.columns):
		pol_df['sr_dist_km']=np.nan

	# Rounds lat/lon to the specified number of decimal places
	if p_dict['epicenter_degree_precision']>0:
		if 'origin_lon' in pol_df.columns:
			pol_df['origin_lon']=pol_df['origin_lon'].round(p_dict['epicenter_degree_precision'])
		if 'origin_lat' in pol_df.columns:
			pol_df['origin_lat']=pol_df['origin_lat'].round(p_dict['epicenter_degree_precision'])

	# Applies default location uncertainties if none were given by the user
	if not('horz_uncert_km' in pol_df.columns):
		pol_df['horz_uncert_km']=p_dict['default_horz_uncert_km']
	if not('vert_uncert_km' in pol_df.columns):
		pol_df['vert_uncert_km']=p_dict['default_vert_uncert_km']

	# Applies default location uncertainties if location uncertainties are <0
	pol_df.loc[ (pd.isnull(pol_df['horz_uncert_km'])) | (pol_df['horz_uncert_km']<0),'horz_uncert_km']=p_dict['default_horz_uncert_km']
	pol_df.loc[ (pd.isnull(pol_df['vert_uncert_km'])) | (pol_df['vert_uncert_km']<0),'vert_uncert_km']=p_dict['default_vert_uncert_km']

	# Flips takeoff notation. This flips takeoff angles (>90: down, <90: up) into HASH notation (<90: upgoing, >90: downgoing)
	if p_dict['flip_takeoff']:
		if not('takeoff' in pol_df.columns):
			raise ValueError('Unable to flip takeoff angle notation. No takeoff angles found.')
		pol_df['takeoff']=180-pol_df['takeoff']

	# Checks to make sure all event_ids in the polarity file(s) are included in the catalog.
	if len(cat_df)>0:
		unique_pol_event_id=pol_df['event_id'].unique()
		id_found_flag=np.isin(unique_pol_event_id,cat_df['event_id'])
		if np.any(~id_found_flag):
			raise ValueError('{} events in polarity or SP files not found in catalog.\n\tMissing problematic IDs: {}'.
				format(np.sum(~id_found_flag),unique_pol_event_id[np.where(~id_found_flag)[0]]))

	# Ensures that there are not null values for the polarities/sp_ratios
	if 'sp_ratio' in pol_df:
		null_flag=pd.isnull(pol_df[['p_polarity','sp_ratio']]).all(axis=1)
		if null_flag.any():
			raise ValueError('Null values are present for both the polarity and S/P columns. Example problematic records:\n{}'.
							format(pol_df.loc[null_flag,['event_id','sta_code','p_polarity','sp_ratio','source']].head(10).to_string()))
	else:
		null_flag=pd.isnull(pol_df['p_polarity'])
		if null_flag.any():
			raise ValueError('Null values are present for the polarity column. Example problematic records:\n{}'.
							format(pol_df.loc[null_flag,['event_id','sta_code','p_polarity','source']].head(10).to_string()))

	# Ensures there aren't null values in the polarity dataset
	null_pol_df=pol_df.drop(pol_df.filter(['sr_dist_km','takeoff','azimuth','takeoff_uncertainty','azimuth_uncertainty','sp_ratio','p_polarity','station']),axis=1)
	if null_pol_df.isnull().values.any():
		null_index=null_pol_df[null_pol_df.isna().any(axis=1)].index
		if len(null_index)>5:
			null_index=null_index[:5]
		raise ValueError('There are null values in datasets. If you are using multiple polarity files, do the columns match?\n{}'.format(null_pol_df.loc[null_index,:].to_string()))

	# Ensures P polarity weights are in the appropriate [-1,1] range
	if (pol_df['p_polarity'].abs()>1).any():
		tmp_ind=pol_df['p_polarity'].abs().argmax()
		raise ValueError('P polarity weights should range from -1 to 1. Example problem polarity:\n{}'.format(pol_df.loc[[tmp_ind],:]))

	# Discards P polarities with weights below a selected value
	if p_dict['min_polarity_weight']>0:
		drop_flag=pol_df['p_polarity'].abs()<p_dict['min_polarity_weight']
		if any(drop_flag):
			print('Discarded {} polarity measurements with weights less than {}'.format(sum(drop_flag),p_dict['min_polarity_weight']))
			pol_df=pol_df.drop(pol_df[drop_flag].index).reset_index(drop=True)

	# If takeoff and azimuth columns exist but the uncertainties are missing, we'll use the default uncertainty values
	if {'takeoff','azimuth'}.issubset(pol_df.columns):
		if not('takeoff_uncertainty' in pol_df.columns):
			if p_dict['default_takeoff_uncert']>0:
				print('No precomputed takeoff uncertainties provided. Applying {} deg takeoff uncertainty'.format(p_dict['default_takeoff_uncert']))
				pol_df['takeoff_uncertainty']=p_dict['default_takeoff_uncert']
			else:
				print('*WARNING: No precomputed takeoff uncertainty provided, and default_takeoff_uncert==0. Takeoffs will not be perturbed.')
				pol_df['takeoff_uncertainty']=0
		if not('azimuth_uncertainty' in pol_df.columns):
			if p_dict['default_azimuth_uncert']>0:
				print('No precomputed azimuth uncertainties provided. Applying {} deg azimuthual uncertainty'.format(p_dict['default_azimuth_uncert']))
				pol_df['default_azimuth_uncert']=p_dict['default_azimuth_uncert']
			else:
				print('*WARNING: No precomputed azimuth uncertainties provided, and default_takeoff_uncert==0. Source-receiver azimuths will not be perturbed.')
				pol_df['azimuth_uncertainty']=0

	# If either takeoff and azimuth columns exist (but not both), ignore the provided column
	if (pd.isnull(pol_df['azimuth']).any()) is not (pd.isnull(pol_df['takeoff']).any()):
		raise ValueError('Complete azimuth and takeoff angle information should be provided when using precomputed ray info.')

	if 'takeoff_uncertainty' in pol_df.columns:
		takeoff_null_flag=pd.isnull(pol_df['takeoff_uncertainty'])
		if takeoff_null_flag.all():
			pol_df['takeoff_uncertainty']=p_dict['default_takeoff_uncert']
		elif takeoff_null_flag.any():
			print('Some takeoff uncertainties are missing. Applying {} deg takeoff uncertainty'.format(p_dict['default_takeoff_uncert']))
			pol_df.loc[takeoff_null_flag,'takeoff_uncertainty']=p_dict['default_takeoff_uncert']

	if 'azimuth_uncertainty' in pol_df.columns:
		azimuth_null_flag=pd.isnull(pol_df['azimuth_uncertainty'])
		if azimuth_null_flag.all():
			pol_df['azimuth_uncertainty']=p_dict['default_azimuth_uncert']
		elif azimuth_null_flag.any():
			print('Some azimuth uncertainties are missing. Applying {} deg azimuth uncertainty.'.format(p_dict['default_azimuth_uncert']))
			pol_df.loc[azimuth_null_flag,'azimuth_uncertainty']=p_dict['default_azimuth_uncert']

	# Determines if takeoff/azimuths need (or can) be calculated
	null_index=pol_df[pol_df.loc[:,['takeoff','azimuth','takeoff_uncertainty','azimuth_uncertainty']].isnull().any(axis=1)].index
	if (not(p_dict['vmodel_paths'] and p_dict['stfile'])) | (null_index.empty):
		p_dict['compute_takeoff_azimuth']=False
		if not(null_index.empty):
			print('*WARNING: {} of of the {} polarities have missing takeoff/azimuths. These polarities will be ignored.'.format(len(null_index),len(pol_df)))
			pol_df=pol_df.drop(null_index).reset_index(drop=True)

	return pol_df


def check_model_eq(cat_df,pol_df,p_dict):
	'''
	Checks that earthquake locations will be contained in the model
	'''
	# Ensures the depth range is on the interval p_dict['look_dep'][2]
	if ((p_dict['look_dep'][1]-p_dict['look_dep'][0]) % p_dict['look_dep'][2])!=0:
		new_dep2=p_dict['look_dep'][0]+(np.ceil((p_dict['look_dep'][1]-p_dict['look_dep'][0])/p_dict['look_dep'][2]))*p_dict['look_dep'][2]
		print('Depth range ({}-{} km) for the lookup table is not on the interval {} km. Changing look_dep[1] to {}'.format(p_dict['look_dep'][0],p_dict['look_dep'][1],p_dict['look_dep'][2],new_dep2))
		p_dict['look_dep'][1]=new_dep2

	# Ensures the distance range is on the interval p_dict['look_del'][2]
	if ((p_dict['look_del'][1]-p_dict['look_del'][0]) % p_dict['look_del'][2])!=0:
		new_del2=p_dict['look_del'][0]+(np.ceil((p_dict['look_del'][1]-p_dict['look_del'][0])/p_dict['look_del'][2]))*p_dict['look_del'][2]
		print('Distance range ({}-{} km) for the lookup table is not on the interval {} km. Changing look_del[1] to {}'.
			format(p_dict['look_del'][0],p_dict['look_del'][1],p_dict['look_del'][2],new_del2) )
		p_dict['look_del'][1]=new_del2

	num_source_depth_bins=int((p_dict['look_dep'][1]-p_dict['look_dep'][0])/3+1)
	if num_source_depth_bins>p_dict['nd0']:
		raise ValueError('Given the lookup depth range of {}-{}km with interval {}km look_dep), the {} source depth bins needed exceeds the maximum number of source depth bins {} (nd0).'.format(p_dict['look_dep'][0],p_dict['look_dep'][1],p_dict['look_dep'][2],num_source_depth_bins,p_dict['nd0']))

	if p_dict['delmin']<=0:
		p_dict['delmin']=p_dict['look_del'][0]
	# If the maximum source-receiver distance is not specified, sets it equal to the velocity model extent
	if p_dict['delmax']<=0:
		p_dict['delmax']=p_dict['look_del'][1]

	# Ensures that the catalog earthquake depths are all contained within the modeled interval.
	if len(cat_df)>0:
		if cat_df['origin_depth_km'].max()>p_dict['look_dep'][1]:
			prob_ind=cat_df['origin_depth_km'].argmax()
			if p_dict['allow_hypocenters_outside_table']:
				print('Earthquake depth ({} km) for event {} is greater than maximum lookup table depth ({} km). Setting hypocentral depth to {} km.'.
					format(cat_df['origin_depth_km'].max(),cat_df.loc[prob_ind,'event_id'],p_dict['look_dep'][1],p_dict['look_dep'][1]))
				cat_df.loc[cat_df['origin_depth_km']>p_dict['look_dep'][1],'origin_depth_km']=p_dict['look_dep'][1]
			else:
				raise ValueError('Earthquake depth ({} km) for event {} is greater than maximum lookup table depth ({} km).'.
					format(cat_df['origin_depth_km'].max(),cat_df.loc[prob_ind,'event_id'],p_dict['look_dep'][1]))
		if cat_df['origin_depth_km'].min()<p_dict['look_dep'][0]:
			prob_ind=cat_df['origin_depth_km'].argmin()
			if p_dict['allow_hypocenters_outside_table']:
				print('WARNING: Earthquake depth ({} km) for event {} is less than minimum lookup table depth ({} km). Setting hypocentral depth to {} km.'.
					format(cat_df['origin_depth_km'].min(),cat_df.loc[prob_ind,'event_id'],p_dict['look_dep'][0],p_dict['look_dep'][0]))
				cat_df.loc[cat_df['origin_depth_km']<p_dict['look_dep'][0],'origin_depth_km']=p_dict['look_dep'][0]
			else:
				raise ValueError('Earthquake depth ({} km) for event {} is less than maximum lookup table depth ({} km).'.
					format(cat_df['origin_depth_km'].min(),cat_df.loc[prob_ind,'event_id'],p_dict['look_dep'][0]))
	if 'origin_depth_km' in pol_df.columns:
		if pol_df['origin_depth_km'].max()>p_dict['look_dep'][1]:
			if p_dict['allow_hypocenters_outside_table']:
				print('WARNING: Earthquake depth ({}) in polarity record is greater than the maximum lookup table depth ({} km). Setting hypocentral depth to {} km.'.
					format(pol_df['origin_depth_km'].max(),p_dict['look_dep'][1],p_dict['look_dep'][1]))
				pol_df.loc[pol_df['origin_depth_km']>p_dict['look_dep'][1],'origin_depth_km']=p_dict['look_dep'][1]
			else:
				raise ValueError('Earthquake depth ({}) in polarity record is greater than the maximum lookup table depth ({} km)'.
					format(pol_df['origin_depth_km'].max(),p_dict['look_dep'][1]))
		if pol_df['origin_depth_km'].min()<p_dict['look_dep'][0]:
			if p_dict['allow_hypocenters_outside_table']:
				print('WARNING: Earthquake depth ({}) in polarity record is less than the minimum lookup table depth ({} km). Setting hypocentral depth to {} km.'.
					format(pol_df['origin_depth_km'].min(),p_dict['look_dep'][0],p_dict['look_dep'][0]))
				pol_df.loc[pol_df['origin_depth_km']<p_dict['look_dep'][0],'origin_depth_km']=p_dict['look_dep'][0]
			else:
				raise ValueError('Earthquake depth ({}) in polarity record is less than the maximum lookup table depth ({} km)'.
					format(pol_df['origin_depth_km'].min(),p_dict['look_dep'][0]))

	# Warns the user if a permiated earthquake depth could fall outside of the modeled interval.
	if len(cat_df)>0:
		max_eq_perm_depth=(cat_df['origin_depth_km']+cat_df['vert_uncert_km']).max()
		if max_eq_perm_depth>p_dict['look_dep'][1]:
			prob_ind=np.argmax(max_eq_perm_depth)
			print(('*WARNING: Given the earthquake depths and vertical uncertainties, permiated hypocenters could fall '+
			'outside of the maximum lookup table depth of {} km. Example event: max permiated depth of {} km for event {}.'+
			'You may wish to increase the depth of the lookup table.').format(p_dict['look_dep'][1],max_eq_perm_depth,cat_df.loc[prob_ind,'event_id']))

	return pol_df,p_dict



'''
Functions for plotting beachballs.
This partially uses python code from ObsPy, which adapted code from bb.m
written by Andy Michael, Chen Ji and Oliver Boyd.

bb.m: http://www.ceri.memphis.edu/people/olboyd/Software/Software.html
Obspy: https://docs.obspy.org/packages/autogen/obspy.imaging.beachball.html
'''
# Standard libraries
import copy
import os

# External libraries
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt


def plot_mech(mech_df,pol_df,takeoff,azimuth,p_dict,acceptable_sdr=np.zeros([3,0])):
    '''
    Creates a plot of the mechanism with the P-polarity and S/P measurements.
    '''
    if (len(pol_df)==0) | (len(mech_df)==0):
        return

    event_id=str(pol_df.event_id.values[0])
    xy=takeoff_az2xy(takeoff,azimuth)

    up_ind=np.where(pol_df['p_polarity']>0)[0]
    down_ind=np.where(pol_df['p_polarity']<0)[0]

    plot_sp=False
    if 'sp_ratio' in pol_df.columns:
        sp_ind=np.where(~pd.isnull(pol_df['sp_ratio']))[0]
        if len(sp_ind)>0:
            plot_sp=True

    beach_plots=[]
    if (len(mech_df)==1) or (p_dict['plot_mult_solutions']==False):
        beach_plots=[beach(mech_df.loc[0,'str_avg'],mech_df.loc[0,'dip_avg'],mech_df.loc[0,'rak_avg'],facecolor='0.75',linewidth=0.5,zorder=0)]
    else:
        for mech_ind in range(len(mech_df)):
            beach_plots.append(beach(mech_df.loc[mech_ind,'str_avg'],mech_df.loc[mech_ind,'dip_avg'],mech_df.loc[mech_ind,'rak_avg'],facecolor='.1',linewidth=0.5,zorder=0,alpha=1/len(mech_df)*4/3))

    beach_accept=[]
    if acceptable_sdr.shape[1]:
        for accept_ind in range(acceptable_sdr.shape[1]):
            beach_accept.append(beach(acceptable_sdr[0,accept_ind],acceptable_sdr[1,accept_ind],acceptable_sdr[2,accept_ind],facecolor='None',bgcolor='None',linewidth=0.125,zorder=1,edgecolor='0.5',alpha=0))
            # beach_accept.append(beach(acceptable_sdr[:,accept_ind],acceptable_sdr[:,accept_ind],acceptable_sdr[:,accept_ind],facecolor='None',linewidth=0.5))

    if not(plot_sp): # When plotting only P polarities
        fig,axes=plt.subplots(1,1,figsize=(5,5))
        axes=[axes]
        # axes[0].add_collection(beach1)
        for mech_ind in range(len(beach_plots)):
            axes[0].add_collection(beach_plots[mech_ind])
        axes[0].set_title('\nEvent ID: {}'.format(event_id))
    else: # When plotting S/P ratios too
        fig,axes=plt.subplots(1,2,figsize=(10,5))
        for mech_ind in range(len(beach_plots)):
            axes[0].add_collection(copy.deepcopy(beach_plots[mech_ind]))      
        axes[0].set_title('\nEvent ID: {}, {}'.format(event_id,'P-polarities'))
        axes[1].set_title('\nEvent ID: {}, {}'.format(event_id,'S/P Ratios'))
        for mech_ind in range(len(beach_plots)):
            axes[1].add_collection(copy.deepcopy(beach_plots[mech_ind]))  

        # Plots S/P ratios sized by values
        axes[1].scatter(xy[sp_ind,0],xy[sp_ind,1],s=10**(pol_df['sp_ratio'].values[sp_ind])*10,marker='o',linewidths=.5, edgecolor='k',facecolor='None',zorder=3)

    if beach_accept:
        beach1_2=beach(mech_df.loc[0,'str_avg'],mech_df.loc[0,'dip_avg'],mech_df.loc[0,'rak_avg'],edgecolor='k',linewidth=0.5,facecolor='None',bgcolor='None',zorder=2,alpha=0)
        for ax in axes:
            for beach_acc in beach_accept:
                ax.add_collection(copy.deepcopy(beach_acc))
            ax.add_collection(copy.deepcopy(beach1_2))    
            
            # for accept_ind in range(len(beach_accept)):
            #     axes[0].add_collection(copy.deepcopy(beach_accept[accept_ind]))

    # Plots Up/Down polarities
    if len(up_ind)>0:
        axes[0].scatter(xy[up_ind,0],xy[up_ind,1],marker='+',s=50, linewidths=.6, c='k',zorder=3)
    if len(down_ind)>0:
        axes[0].scatter(xy[down_ind,0],xy[down_ind,1],marker='o',s=50, linewidths=.5, edgecolor='k', facecolor='None',zorder=3)

    if p_dict['plot_station_names']:
        if p_dict['require_network_match']:
            sta_names=pol_df.sta_code.str.split('.').str[1].values
        else:
            sta_names=pol_df.sta_code.str.split('.').str[0].values

        for pick_ind in up_ind:
            axes[0].text(xy[pick_ind,0],xy[pick_ind,1],sta_names[pick_ind],fontsize=6,zorder=4)
        for pick_ind in down_ind:
            axes[0].text(xy[pick_ind,0],xy[pick_ind,1],sta_names[pick_ind],fontsize=6,zorder=4)
        if plot_sp:    
            for pick_ind in sp_ind:
                axes[1].text(xy[pick_ind,0],xy[pick_ind,1],sta_names[pick_ind],fontsize=6,zorder=4)    

    for ax in axes:
        ax.set_xlim([-1.01,1.01])
        ax.set_ylim([-1.01,1.01])
        ax.set_aspect(1)
        ax.set_xticks([])
        ax.set_yticks([])

    fig.tight_layout()
    if not(os.path.exists(p_dict['outfolder_plots'])):
        os.makedirs(p_dict['outfolder_plots'],exist_ok=True)
    plt.savefig(os.path.join(p_dict['outfolder_plots'],str(event_id)+'.png'),dpi=150)
    plt.close()


def takeoff_az2xy(takeoff,azimuth,projection='lambert'):
    '''
    Projects takeoff and azimuths onto focal sphere.
    Supported projections are 'stereographic' and 'lambert'

    Takeoffs:
        >90: downgoing, <90: upgoing
    Azimuths:
        # 0: North, 90: east, etc.
    '''
    takeoff=180-takeoff
    r=np.ones(len(takeoff))
    r[takeoff>90]=-1

    theta=np.deg2rad(takeoff)
    phi=np.deg2rad(90-azimuth)
    xyz=np.empty((3,len(takeoff)),dtype=float)
    xyz[0,:]=r*np.sin(theta)*np.cos(phi)
    xyz[1,:]=r*np.sin(theta)*np.sin(phi)
    xyz[2,:]=r*np.cos(theta)
    if projection=='stereographic':
        xy=xyz[:2,:]/(1+xyz[2,:])
    elif projection=='lambert':
        xy=xyz[:2,:]/np.sqrt(1+xyz[2,:])
    else:
        raise ValueError('Unknown projection: {}'.format(projection))
    return xy.T


def beach(strike, dip, rake, linewidth=.5, facecolor='0.75', bgcolor='w',
            edgecolor='k',alpha=1., nofill=False, zorder=0):
    '''
    Return a beach ball as a collection.

    From ObsPy:
    https://docs.obspy.org/packages/autogen/obspy.imaging.beachball.html
    '''
    from matplotlib import collections

    colors, p = plot_dc(strike,dip,rake)
    col = collections.PatchCollection(p, match_original=False)
    col.set_facecolors([facecolor,bgcolor])

    col.set_edgecolor(edgecolor)
    col.set_linewidth(linewidth)
    col.set_zorder(zorder)
    if alpha>0:
        col.set_alpha(alpha)

    return col


def strike_dip(n, e, u):
    '''
    Finds strike and dip of plane given normal vector having components n, e,
    and u.

    Adapted from MATLAB script
    `bb.m <http://www.ceri.memphis.edu/people/olboyd/Software/Software.html>`_
    written by Andy Michael, Chen Ji and Oliver Boyd.

    From ObsPy:
    https://docs.obspy.org/packages/autogen/obspy.imaging.beachball.html
    '''
    r2d = 180 / np.pi
    if u < 0:
        n = -n
        e = -e
        u = -u

    strike = np.arctan2(e, n) * r2d
    strike = strike - 90
    while strike >= 360:
        strike = strike - 360
    while strike < 0:
        strike = strike + 360
    x = np.sqrt(np.power(n, 2) + np.power(e, 2))
    dip = np.arctan2(x, u) * r2d
    return (strike, dip)


def aux_plane(s1, d1, r1):
    '''
    Get Strike and dip of second plane.

    Adapted from MATLAB script
    `bb.m <http://www.ceri.memphis.edu/people/olboyd/Software/Software.html>`_
    written by Andy Michael, Chen Ji and Oliver Boyd.

    From ObsPy:
    https://docs.obspy.org/packages/autogen/obspy.imaging.beachball.html
    '''
    r2d = 180 / np.pi

    z = (s1 + 90) / r2d
    z2 = d1 / r2d
    z3 = r1 / r2d
    # slick vector in plane 1
    sl1 = -np.cos(z3) * np.cos(z) - np.sin(z3) * np.sin(z) * np.cos(z2)
    sl2 = np.cos(z3) * np.sin(z) - np.sin(z3) * np.cos(z) * np.cos(z2)
    sl3 = np.sin(z3) * np.sin(z2)
    (strike, dip) = strike_dip(sl2, sl1, sl3)

    n1 = np.sin(z) * np.sin(z2)  # normal vector to plane 1
    n2 = np.cos(z) * np.sin(z2)
    h1 = -sl2  # strike vector of plane 2
    h2 = sl1

    z = h1 * n1 + h2 * n2
    z = z / np.sqrt(h1 * h1 + h2 * h2)
    # we might get above 1.0 only due to floating point
    # precision. Clip for those cases.
    float64epsilon = 2.2204460492503131e-16
    if 1.0 < abs(z) < 1.0 + 100 * float64epsilon:
        z = np.copysign(1.0, z)
    z = np.arccos(z)
    rake = 0
    if sl3 > 0:
        rake = z * r2d
    if sl3 <= 0:
        rake = -z * r2d
    return (strike, dip, rake)


def pol2cart(th, r):
    '''
    From ObsPy:
    https://docs.obspy.org/packages/autogen/obspy.imaging.beachball.html
    '''
    x = r * np.cos(th)
    y = r * np.sin(th)
    return (x, y)


def xy2patch(x, y, res, xy):
    '''
    From ObsPy:
    https://docs.obspy.org/packages/autogen/obspy.imaging.beachball.html
    '''
    # check if one or two resolutions are specified (Circle or Ellipse)
    from matplotlib import path as mplpath
    from matplotlib import patches
    try:
        assert len(res) == 2
    except TypeError:
        res = (res, res)
    # transform into the Path coordinate system
    x = x * res[0] + xy[0]
    y = y * res[1] + xy[1]
    verts = list(zip(x.tolist(), y.tolist()))
    codes = [mplpath.Path.MOVETO]
    codes.extend([mplpath.Path.LINETO] * (len(x) - 2))
    codes.append(mplpath.Path.CLOSEPOLY)
    path = mplpath.Path(verts, codes)
    return patches.PathPatch(path)


def plot_dc(strike,dip,rake, size=100, xy=(0, 0), width=2):
    '''
    Uses one nodal plane of a double couple to draw a beach ball plot.

    :param ax: axis object of a matplotlib figure
    :param np1: :class:`~NodalPlane`

    Adapted from MATLAB script
    `bb.m <http://www.ceri.memphis.edu/people/olboyd/Software/Software.html>`_
    written by Andy Michael, Chen Ji and Oliver Boyd.

    From ObsPy:
    https://docs.obspy.org/packages/autogen/obspy.imaging.beachball.html
    '''
    # check if one or two widths are specified (Circle or Ellipse)
    try:
        assert len(width) == 2
    except TypeError:
        width = (width, width)
    s_1 = strike
    d_1 = dip
    r_1 = rake
    D2R = np.pi / 180

    m = 0
    if r_1 > 180:
        r_1 -= 180
        m = 1
    if r_1 < 0:
        r_1 += 180
        m = 1

    # Get azimuth and dip of second plane
    (s_2, d_2, _r_2) = aux_plane(s_1, d_1, r_1)

    d = size / 2

    if d_1 >= 90:
        d_1 = 89.9999
    if d_2 >= 90:
        d_2 = 89.9999

    # arange checked for numerical stability, np.pi is not multiple of 0.1
    phi = np.arange(0, np.pi, .01)
    l1 = np.sqrt(
        np.power(90 - d_1, 2) / (
            np.power(np.sin(phi), 2) +
            np.power(np.cos(phi), 2) *
            np.power(90 - d_1, 2) / np.power(90, 2)))
    l2 = np.sqrt(
        np.power(90 - d_2, 2) / (
            np.power(np.sin(phi), 2) + np.power(np.cos(phi), 2) *
            np.power(90 - d_2, 2) / np.power(90, 2)))

    collect = []
    # plot paths, once for tension areas and once for pressure areas
    for m_ in ((m + 1) % 2, m):
        inc = 1
        (x_1, y_1) = pol2cart(phi + s_1 * D2R, l1)

        if m_ == 1:
            lo = s_1 - 180
            hi = s_2
            if lo > hi:
                inc = -1
            th1 = np.arange(s_1 - 180, s_2, inc)
            (xs_1, ys_1) = pol2cart(th1 * D2R, 90 * np.ones((1, len(th1))))
            (x_2, y_2) = pol2cart(phi + s_2 * D2R, l2)
            th2 = np.arange(s_2 + 180, s_1, -inc)
        else:
            hi = s_1 - 180
            lo = s_2 - 180
            if lo > hi:
                inc = -1
            th1 = np.arange(hi, lo, -inc)
            (xs_1, ys_1) = pol2cart(th1 * D2R, 90 * np.ones((1, len(th1))))
            (x_2, y_2) = pol2cart(phi + s_2 * D2R, l2)
            x_2 = x_2[::-1]
            y_2 = y_2[::-1]
            th2 = np.arange(s_2, s_1, inc)
        (xs_2, ys_2) = pol2cart(th2 * D2R, 90 * np.ones((1, len(th2))))
        x = np.concatenate((x_1, xs_1[0], x_2, xs_2[0]))
        y = np.concatenate((y_1, ys_1[0], y_2, ys_2[0]))

        x = x * d / 90
        y = y * d / 90

        # calculate resolution
        res = [value / float(size) for value in width]

        # construct the patch
        collect.append(xy2patch(y, x, res, xy))
    return ['b', 'w'], collect



'''
Functions for creating the fortran gridsearch module.
'''

# Standard libraries
import os
import glob
import sys
from pathlib import Path

# External libraries
import numpy as np
from numpy import f2py


def create():
    '''
    Gives the user the option to automatically create a Python C/API extension
    module for the gridsearch algorithm.

    If the module is created, the user should re-run SKHASH and the module
    should be automatically used.
    '''
    usr_input=input('The Fortran subroutine module does not exist, but you have \'use_fortran=True\'. Would you like to create this module? [y / n]\n')
    usr_input=usr_input.strip().lower()
    if (usr_input!='y') & (usr_input!='yes'):
        print('\nThe module will not be created. If you would like to create the module manually, the following may help.')
        print('\tStep 1. Navigate to the SKHASH/functions folder, e.g.:\n\t\tcd functions')
        print('\tStep 2. Run the following command:\n\t\tpython3 -m numpy.f2py -c gridsearch.f -m gridsearch')
        return 0
    else:
        print('\nCreating module. This may take a minute...')
        gridsearchf_path = Path(__file__).parent / 'gridsearch.f'
        with open(gridsearchf_path) as sourcefile:
            sourcecode = sourcefile.read()
        result=f2py.compile(sourcecode, modulename='gridsearch')
        if result==0:
            so_filepaths=glob.glob('gridsearch.cpython-*-*.so')
            for so_filepath in so_filepaths:
                # os.rename(so_filepath,os.path.join('functions',so_filepath))
                new_so_filepath=Path(__file__).parent / os.path.basename(so_filepath)
                os.rename(so_filepath,new_so_filepath)
            try:
                # import functions.gridsearch as gridsearch
                import gridsearch
                print('\n\nThe fortran extension module has been successfully created!\n')
                return 1
            except Exception as e:
                raise ValueError('There was an issue loading the module. Error message:\n{}'.format(e))
        else:
            if ((sys.version_info.major==3) & (sys.version_info.minor>=12)):
                raise ValueError('The compiled extension module could not be created. Have you installed the \'meson\' and \'ninja\' packages? (pip3 install meson ninja)')
            else:
                raise ValueError('The compiled extension module could not be created.')


def prep_subroutine(sr_azimuth,takeoff,p_pol,sp_amp,npick0,nmc):
    '''
    Gets the polarities, S/P ratios, azimuth, and takeoff information in the
    correct format to call the Fortan subroutine.
    '''
    p_azi_mc=np.zeros((npick0,500))
    p_the_mc=np.zeros((npick0,500))
    p_azi_mc[:sr_azimuth.shape[0],:nmc]=sr_azimuth
    p_the_mc[:sr_azimuth.shape[0],:nmc]=takeoff

    npsta=sr_azimuth.shape[0]

    f_sp_amp=np.zeros(npsta)
    f_sp_amp[:len(sp_amp)]=sp_amp
    f_sp_amp[np.isnan(f_sp_amp)]=-9

    f_p_pol=np.zeros(npsta,float)
    f_p_pol[:len(p_pol)]=p_pol
    p_qual=np.abs(f_p_pol).astype(float)
    f_p_pol[f_p_pol<0]=-1
    f_p_pol[f_p_pol>0]=1
    f_p_pol=f_p_pol.astype(int)

    return p_azi_mc,p_the_mc,f_sp_amp,f_p_pol,p_qual




# Standard libraries
import time

# External libraries
import numpy as np

import functions.gridsearch_so as gridsearch_so # For creating fortran gridsearch module
import functions.out as out # Output functions
import functions.fun as fun # Computing mechanisms


def compute_mech(event_x,num_events,event_id,event_pol_df,p_dict,lookup_dict,qual_criteria_dict,cat_df,dir_cos_dict):
	'''
	Computes focal mechanisms.
	Input:
		event_x: The event number (cosmetic)
		num_events: Total number of events to compute mechanisms (cosmetic)
		event_id: event id string for the event
		event_pol_df: polarity dataframe
		p_dict: Parameter values created in SKHASH.py, dictionary
		lookup_dict: dictionary with lookup variables, produced by create_lookup_table()
		qual_criteria_dict: dictionary of quality criteria, created in SKHASH.py
		cat_df: catalog dataframe
		dir_cos_dict: dictionary of coordinate transformation variables, created by dir_cos_setup()
	Output:
		mech_dict: dictionary of mechanism solutions
	'''
	event_runtime_start = time.time()

	mech_dict={'event_index':-1,'pol_agreement_out':[],
			'sp_diff_out':[],'takeoff':-1.,'sr_az':-1.,
			'takeoff_uncertainty':-1.,'azimuth_uncertainty':-1.,
			'mech_qual':''}

	if p_dict['stfile']: # Perturb earthquake locations and determine azimuth and takeoff angles
		perturbed_origin_depth_km,sr_dist_km,sr_azimuth=fun.perturb_eq_locations(event_pol_df,p_dict['look_dep'],p_dict['perturb_epicentral_location'],nmc=p_dict['nmc'])
		takeoff=fun.lookup_takeoff(lookup_dict['table'],perturbed_origin_depth_km,sr_dist_km,p_dict['look_dep'],p_dict['look_del'],lookup_dict['deptab'],lookup_dict['delttab'],num_velocity_models=len(p_dict['vmodel_paths']))

		# Discards any measurements with a takeoff uncertainty > pmax
		if p_dict['pmax']>0:
			calc_takeoff_uncertainty=np.std(takeoff,axis=1)
			rm_ind=np.where(calc_takeoff_uncertainty>p_dict['pmax'])[0]

			event_pol_df=event_pol_df.drop(event_pol_df.index[rm_ind])
			sr_azimuth=np.delete(sr_azimuth,rm_ind,axis=0)
			takeoff=np.delete(takeoff,rm_ind,axis=0)
			print('Dropped {} measurements for high takeoff uncertainties'.format(len(rm_ind)))
	else: # Perturb predetermined azimuth and takeoff angles
		sr_azimuth,takeoff=fun.perturb_azimuth_takeoff(event_pol_df,p_dict['nmc'])

	# Calculates the maximum azimuthal and takeoff angle gaps, skipping the event if necessary
	max_azimuthal_gap,max_takeoff_gap=fun.determine_max_gap(sr_azimuth[:,0],takeoff[:,0])

	if max_azimuthal_gap>p_dict['max_agap']:
		print('{} / {}\t({})\n'.format(event_x,num_events-1,event_id)+
			  '\tMaximum azimuthal gap ({}) > max_agap ({}). Skipping.'.format(max_azimuthal_gap.round(3),p_dict['max_agap']))
		return mech_dict
	if max_takeoff_gap>p_dict['max_pgap']:
		print('{} / {}\t({})\n'.format(event_x,num_events-1,event_id)+
			  '\tMaximum takeoff angle gap ({}) > max_pgap ({}). Skipping.'.format(max_takeoff_gap.round(3),p_dict['max_pgap']))
		return mech_dict

	# P-polarity parameters for determining best-fit solutions
	p_pol=event_pol_df['p_polarity'].values
	p_pol[~np.isfinite(p_pol)]=0
	sumpolweight=np.sum(np.abs(p_pol))
	nextra=max([round(sumpolweight*p_dict['badfrac']*0.5),p_dict['badmin']])
	ntotal=max([round(sumpolweight*p_dict['badfrac']),p_dict['badmin']])

	# S/P ratio parameters for determining best-fit solutions
	if p_dict['ampfile'] or p_dict['relampfile']:
		sp_amp=event_pol_df['sp_ratio'].values
		nspr=sum(np.isfinite(sp_amp))

		# additional amplitude misfit allowed above minimum
		qextra=max([nspr*p_dict['qbadfrac']*0.5,p_dict['qbadmin']])

		# total allowed amplitude misfit
		qtotal=max([nspr*p_dict['qbadfrac'],p_dict['qbadmin']])
	else:
		sp_amp=np.empty(len(event_pol_df));sp_amp[:]=np.nan
		qextra=0
		qtotal=0

	# Runs the gridsearch to find potential mech solutions.
	if p_dict['use_fortran']: # Uses the Python C/API to call Fortran subroutine
		import functions.gridsearch as gridsearch
		p_azi_mc,p_the_mc,f_sp_amp,f_p_pol,p_qual=gridsearch_so.prep_subroutine(sr_azimuth,takeoff,p_pol,sp_amp,p_dict['npick0'],p_dict['nmc'])
		nf,strike_all,dip_all,rake_all,faultnorms_all,faultslips_all=gridsearch.focalamp_mc_wt(p_azi_mc,p_the_mc,f_sp_amp,f_p_pol,p_qual,p_dict['nmc'],p_dict['dang'],p_dict['maxout'],nextra,ntotal,qextra,qtotal,p_dict['min_amp'],len(f_p_pol))
		faultnorms_all=faultnorms_all[:,:nf]
		faultslips_all=faultslips_all[:,:nf]
		strike_all=strike_all[:nf]
		dip_all=dip_all[:nf]
		rake_all=rake_all[:nf]
	else: # Python version
		faultnorms_all,faultslips_all=fun.focal_gridsearch(sr_azimuth,takeoff,p_pol,sp_amp,dir_cos_dict,nextra,ntotal,qextra,qtotal,p_dict['maxout'],dir_cos_dict['ncoor'])

		# Calculates strike,dip,rake from normal,slip vectors for output
		if ((len(p_dict['outfile2'])>0) | (p_dict['plot_acceptable_solutions'])):
			strike_all,dip_all,rake_all=fun.sdr_from_vector(faultnorms_all,faultslips_all)

	if (faultnorms_all).shape[1]==0:
		print('{} / {}\t({})\n'.format(event_x,num_events-1,event_id)+
			  '\tNo solution found for {}'.format(event_id))
		# return 0
		return mech_dict

	p_the_mc=takeoff[:,0]
	p_azi_mc=sr_azimuth[:,0]

	# Calculates the probabilities for potential mech solutions
	mech_df=fun.mech_probability(faultnorms_all,faultslips_all,p_dict['cangle'],p_dict['prob_max'],iterative_avg=p_dict['iterative_avg'])

	if len(mech_df)==0: # No accepted solution
		print('{} / {}\t({})\n'.format(event_x,num_events-1,event_id)+
			  '\tNo accepted solution found for {}'.format(event_id))
		# return 0
		return mech_dict

	# Calculates the misfit for prefered solutions
	mech_df,pol_agreement_out,sp_diff_out=fun.mech_misfit(mech_df,p_azi_mc,p_the_mc,p_pol,sp_amp)

	mech_df['num_p_pol']=np.sum(p_pol!=0)
	mech_df['num_sp_ratios']=np.sum(~np.isnan(sp_amp))
	mech_df['azimuthal_gap']=np.round(max_azimuthal_gap,p_dict['output_angle_precision'])
	mech_df['takeoff_gap']=np.round(max_takeoff_gap,p_dict['output_angle_precision'])

	# Mech solution quality rating
	mech_df=fun.mech_quality(mech_df,qual_criteria_dict)

	if p_dict['min_quality_report']:
		mech_df=mech_df.loc[mech_df['qual']<=p_dict['min_quality_report'],:].reset_index(drop=True)
		if len(mech_df)==0: # No accepted solution
			print('{} / {}\t({})\n'.format(event_x,num_events-1,event_id)+
				'\tNo solution met the minimum quality ({}) for {}'.format(p_dict['min_quality_report'],event_id))
			return mech_dict

	# Rounds mech solution values
	angle_col=['str_avg', 'dip_avg', 'rak_avg','rms_diff','rms_diff_aux']
	quality_col=['prob','mfrac','mavg','stdr']
	mech_df[angle_col]=mech_df[angle_col].round(p_dict['output_angle_precision'])
	mech_df[quality_col]=(mech_df[quality_col]*100).round(p_dict['output_quality_precision'])

	# Writes preferred mechanisms to file
	if p_dict['outfile1']:
		if len(cat_df):
			event_df=cat_df.loc[cat_df.event_id==event_id].iloc[0]
		else:
			event_df=[]
		out.write_outfile1(p_dict['outfile1'],mech_df,event_df,event_id)

	# Writes acceptable mechanisms to file
	if p_dict['outfile2']:
		out.write_outfile2(p_dict['outfile2'],event_id,strike_all,dip_all,rake_all,faultnorms_all,faultslips_all,p_dict['output_angle_precision'],p_dict['output_vector_precision'])

	if (p_dict['stfile']) and (p_dict['outfile_pol_info']):
		out_takeoff=np.round(p_the_mc,p_dict['output_angle_precision'])
		out_sr_az=np.round(p_azi_mc,p_dict['output_angle_precision'])
	else:
		out_takeoff=-999
		out_sr_az=-999

	# Plots the focal mechanism and saves the figure
	if p_dict['outfolder_plots']:
		import functions.plot_mech as plot_mech # For plotting mechanism solutions
		if p_dict['plot_acceptable_solutions']:
			plot_mech.plot_mech(mech_df,event_pol_df,takeoff[:,0],sr_azimuth[:,0],p_dict,acceptable_sdr=np.vstack([strike_all,dip_all,rake_all]))
		else:
			plot_mech.plot_mech(mech_df,event_pol_df,takeoff[:,0],sr_azimuth[:,0],p_dict)

	# Calculates the stdev of the takeoff and azimuths
	if p_dict['outfile_pol_info'] and p_dict['stfile']:
		takeoff_uncertainty_out=np.std(takeoff,axis=1).round(p_dict['output_angle_precision'])

		rad_sr_azimuth=np.deg2rad(sr_azimuth)
		sr_sin_mean=np.mean(np.sin(rad_sr_azimuth),axis=1)
		sr_cos_mean=np.mean(np.cos(rad_sr_azimuth),axis=1)
		R=np.minimum(1,np.hypot(sr_sin_mean, sr_cos_mean))
		azimuth_uncertainty_out=np.degrees(np.sqrt(-2*np.log(R))).round(p_dict['output_angle_precision'])
	else:
		takeoff_uncertainty_out=-1.
		azimuth_uncertainty_out=-1.

	# Prints summary results for the event to std out
	print('{} / {}\t({})\n\tS: {}   D: {}   R: {}   U: {}   Q: {}\n\tRuntime: {:.2f} sec'.format(
					event_x,
					num_events-1,
					event_id,
					mech_df.loc[0,'str_avg'],
					mech_df.loc[0,'dip_avg'],
					mech_df.loc[0,'rak_avg'],
					mech_df.loc[0,'rms_diff'],
					mech_df.loc[0,'qual'],
					(time.time()-event_runtime_start) ),flush=True)

	# return event_pol_df
	return {'event_index':event_pol_df.index,'pol_agreement_out':pol_agreement_out,
			'sp_diff_out':sp_diff_out,'takeoff':out_takeoff,'sr_az':out_sr_az,
			'takeoff_uncertainty':takeoff_uncertainty_out,'azimuth_uncertainty':azimuth_uncertainty_out,
			'mech_qual':mech_df.loc[0,'qual']}



'''
Functions for reading P-polarity inputs.
'''

# External libraries
import numpy as np
import pandas as pd


def read_polarity_file(fpfile,input_format,merge_on):
	if (input_format=='skhash'):
		tmp_pol_df=read_skhash_polarity_file(fpfile,merge_on)
		cat_df=[]
	elif (input_format=='ncsn') or (input_format=='hypoinverse'):
		cat_df,tmp_pol_df=read_ncsn_polarity_file(fpfile)
	elif input_format=='hash1':
		cat_df,tmp_pol_df=read_hash1_polarity_file(fpfile)
	elif (input_format=='hash2') | (input_format=='hash3') | (input_format=='hash5'):
		cat_df,tmp_pol_df=read_hash3_polarity_file(fpfile)
	elif (input_format=='hash4'):
		cat_df,tmp_pol_df=read_hash4_polarity_file(fpfile)
	elif (input_format=='quakeml'):
		cat_df,tmp_pol_df=read_quakeml_polarity_file(fpfile)
	else:
		raise ValueError('Unknown p-wave first motion polarity file format (fpfile: {}).'.format(input_format))

	# Ensures column data formats are the expected type
	tmp_pol_df['event_id']=tmp_pol_df['event_id'].astype(str)
	tmp_pol_df['p_polarity']=tmp_pol_df['p_polarity'].astype(float)
	for merge_on_it in merge_on:
		tmp_pol_df[merge_on_it]=tmp_pol_df[merge_on_it].astype(str)
	if 'event_id2' in tmp_pol_df.columns:
		tmp_pol_df['event_id2']=tmp_pol_df['event_id2'].astype(str)
	if 'sr_dist_km' in tmp_pol_df.columns:
		tmp_pol_df['sr_dist_km']=tmp_pol_df['sr_dist_km'].astype(float)
	if 'takeoff' in tmp_pol_df.columns:
		tmp_pol_df['takeoff']=tmp_pol_df['takeoff'].astype(float)
	if 'azimuth' in tmp_pol_df.columns:
		tmp_pol_df['azimuth']=tmp_pol_df['azimuth'].astype(float)
	if 'takeoff_uncertainty' in tmp_pol_df.columns:
		tmp_pol_df['takeoff_uncertainty']=tmp_pol_df['takeoff_uncertainty'].astype(float)
	if 'azimuth_uncertainty' in tmp_pol_df.columns:
		tmp_pol_df['azimuth_uncertainty']=tmp_pol_df['azimuth_uncertainty'].astype(float)

	# Computes "sta codes" by combining the desired metadata (network, station, location, channel) codes
	if len(tmp_pol_df):
		tmp_pol_df['sta_code']=tmp_pol_df[merge_on].agg('.'.join, axis=1)
	else:
		tmp_pol_df=tmp_pol_df.join(pd.DataFrame(columns=['sta_code']))

	# List of desired columns. We have to keep station codes for HASH simulps and polarity reversal compatability
	require_col=['event_id','sta_code','p_polarity','station','sr_dist_km','takeoff','azimuth','takeoff_uncertainty','azimuth_uncertainty']
	additional_col=['event_id2','origin_lat','origin_lon','origin_depth_km']

	# Selects only the desired columns.
	tmp_pol_df=tmp_pol_df.filter(np.hstack((require_col,additional_col)))

	# Ensures all desired columns exist
	for col in require_col:
		if col not in tmp_pol_df.columns:
			tmp_pol_df[col]=np.nan

	return cat_df,tmp_pol_df


def read_skhash_polarity_file(fpfile,merge_on):
	'''
	Reads polarity file using the SKHASH format.
	'''
	consider_cols=['event_id','event_id2','network','station','location','channel','p_polarity','takeoff','takeoff_uncertainty','azimuth','azimuth_uncertainty','sr_dist_km','origin_latitude','origin_longitude','origin_depth_km','horz_uncert_km','vert_uncert_km']
	try:
		tmp_pol_df=pd.read_csv(fpfile,skipinitialspace=True,comment='#',usecols=lambda x: x in consider_cols)
	except pd.errors.EmptyDataError:
		raise ValueError('fpfile ({}) is empty.'.format(fpfile))
	if not({'event_id', 'p_polarity'}.issubset(tmp_pol_df.columns)):
		raise ValueError('When using SKHASH input format, the fpfile ({}) must contain the columns "event_id" and "p_polarity".'.format(fpfile))
	for req_col in merge_on:
		if not(req_col in tmp_pol_df.columns):
			raise ValueError('When using SKHASH input format and require_{}_match==True, the fpfile ({}) must contain the column "{}".'.format(req_col,fpfile,req_col))

	tmp_pol_df=tmp_pol_df.rename(columns={'origin_latitude':'origin_lat','origin_longitude':'origin_lon'})

	return tmp_pol_df


def dm2dd(deg, min, hemisphere):
	'''
	Converts degrees-minutes to decimal degrees
	'''
	dd = float(deg) + float(min)/60
	if hemisphere == 'W' or hemisphere == 'S':
		dd *= -1
	return dd


def read_ncsn_polarity_file(fpfile,use_weight_code=True,p_weight_I=1.0,p_weight_E=0.5):
	'''
	Reads a USGS NCSN catalog + phase polarity file in hypoinverse format.
	Polarities can be obtained from the NCEDC (https://ncedc.org/ncedc/catalog-search.html)
	and selecting "NCSN catalog + Phase in Hypoinverse format".

	If use_weight_code==False:
		Gives Impulsive (I) manual picks a weight of p_weight_I (default 1.0)
		Gives Emmergent (E) manual picks a weight of p_weight_E (default 0.5)
		If no I/E flag is provided, assumes E.
	If use_weight_code==True:
		pick_df.loc[pick_df['p_weight_code']==0,'p_polarity']=1
		pick_df.loc[pick_df['p_weight_code']==1,'p_polarity']=0.5
		pick_df.loc[pick_df['p_weight_code']==2,'p_polarity']=0.2
		pick_df.loc[pick_df['p_weight_code']==3,'p_polarity']=0.1
		(see https://ncedc.org/pub/doc/ncsn/shadow2000.pdf) for the weight of the polarity.
	'''
	col_ind_name=np.asarray([
		[0,4,'year'],
		[4,6,'month'],
		[6,8,'day'],
		[8,10,'hour'],
		[10,12,'minute'],
		[12,16,'second'],
		[16,18,'lat_deg'],
		[18,19,'latNS'],
		[19,23,'lat_min'],
		[23,26,'lon_deg'],
		[26,27,'lonEW'],
		[27,31,'lon_min'],
		[31,36,'depth'],
		[85,89,'x_error_km'],
		[89,93,'z_error_km'],
		[136,146,'event_id'],
		[147,150,'mag_pref']
	])
	col_ind=col_ind_name[:,0:2].astype(int)
	col_name=col_ind_name[:,2]

	pick_col_ind_name=np.asarray([
		[0,5,'station'],
		[5,7,'network'],
		[9,12,'channel'],
		[13,14,'p_onset'],
		[15,16,'p_first_motion'],
		[16,17,'p_weight_code'],
		[111,113,'location'],
	])
	pick_col_ind=pick_col_ind_name[:,0:2].astype(int)
	pick_col_name=pick_col_ind_name[:,2]

	header_all=[]
	pick_all=[]
	with open(fpfile, 'r') as file1:
		for line in file1:
			if line[0:3]=='   ': # Footer line
				tmp_pick_df=pd.DataFrame(tmp_pol,columns=pick_col_name)

				# Drops any phase arrivals that do not have a P first motion
				tmp_pick_df=tmp_pick_df.drop(tmp_pick_df.loc[tmp_pick_df['p_first_motion']==' '].index)
				# Adds event_id to the dataframe
				tmp_pick_df['event_id']=str(event_id)

				pick_all.append(tmp_pick_df)
			elif (' ' in line[:10]): # Phase arrival line
				tmp_pol.append([line[i:j] for i,j in zip(pick_col_ind[:,0], pick_col_ind[:,1])])
			else: # Header line
				header_all.append([line[i:j] for i,j in zip(col_ind[:,0], col_ind[:,1])])
				header_all[-1][-2]=header_all[-1][-2].strip()
				event_id=header_all[-1][-2]
				tmp_pol=[]
	cat_df=pd.DataFrame(header_all,columns=col_name)
	cat_df=cat_df.astype({'lat_deg':'int32','lat_min':'int32',
								'lon_deg':'int32','lon_min':'int32',
								'depth':'int32','x_error_km':'int32',
								'z_error_km':'int32','mag_pref':'int32',
								'year':'int32','month':'int32',
								'day':'int32','hour':'int32',
								'minute':'int32','second':'float',
						})

	cat_df['second']=cat_df['second']/100
	cat_df['origin_DateTime']=pd.to_datetime(cat_df.loc[:,['year','month','day','hour','minute','second']])
	cat_df['origin_lat']=cat_df['lat_deg']+cat_df['lat_min']/6000
	cat_df.loc[cat_df['latNS']=='S','origin_lat']*=-1
	cat_df['origin_lat']=cat_df['origin_lat'].round(6)
	cat_df['origin_lon']=cat_df['lon_deg']+cat_df['lon_min']/6000
	cat_df.loc[cat_df['lonEW']!='E','origin_lon']*=-1
	cat_df['origin_lon']=cat_df['origin_lon'].round(6)
	cat_df['origin_depth_km']=cat_df['depth']/100
	cat_df['horz_uncert_km']=cat_df['x_error_km']/100
	cat_df['vert_uncert_km']=cat_df['z_error_km']/100
	cat_df['event_mag']=cat_df['mag_pref']/100

	cat_df=cat_df[['origin_DateTime', 'origin_lat', 'origin_lon', 'origin_depth_km',
						'horz_uncert_km', 'vert_uncert_km', 'event_mag', 'event_id']]

	pick_df=pd.concat(pick_all).reset_index(drop=True)
	pick_df['p_polarity']=0.

	if use_weight_code:
		pick_df['p_weight_code']=pick_df['p_weight_code'].astype(int)
		pick_df.loc[pick_df['p_weight_code']==0,'p_polarity']=1
		pick_df.loc[pick_df['p_weight_code']==1,'p_polarity']=0.5
		pick_df.loc[pick_df['p_weight_code']==2,'p_polarity']=0.2
		pick_df.loc[pick_df['p_weight_code']==3,'p_polarity']=0.1
	else:
		pick_df.loc[pick_df['p_onset']=='I','p_polarity']=p_weight_I
		pick_df.loc[(pick_df['p_onset']=='E') | (pick_df['p_onset']==' '),'p_polarity']=p_weight_E

	pick_df.loc[pick_df['p_first_motion']=='D','p_polarity']*=-1
	pick_df=pick_df.drop(columns=['p_first_motion','p_weight_code','p_onset'])

	pick_df['station']=pick_df['station'].str.strip()
	pick_df['network']=pick_df['network'].str.strip()
	pick_df['channel']=pick_df['channel'].str.strip()
	pick_df['location']=pick_df['location'].str.strip()

	return cat_df,pick_df


def read_hash1_polarity_file(fpfile,p_weight_I=1.0,p_weight_E=0.5):
	'''
	Reads a polarity file following HASH driver 1 format.
	Gives Impulsive (I) manual picks a weight of p_weight_I (default 1.0)
	Gives Emmergent (E) manual picks a weight of p_weight_E (default 0.5)
	If no I/E flag is provided, assumes E.
	'''
	col_ind_name=np.asarray([
		[0,2,'year'],
		[2,4,'month'],
		[4,6,'day'],
		[6,8,'hour'],
		[8,10,'minute'],
		[10,14,'second'],
		[14,16,'lat_deg'],
		[16,17,'latNS'],
		[17,21,'lat_min'],
		[21,24,'lon_deg'],
		[24,25,'lonEW'],
		[25,29,'lon_min'],
		[29,34,'depth'],
		[34,36,'mag_pref'],
		[80,84,'x_error_km'],
		[84,88,'z_error_km'],
		[122,138,'event_id'],
	])
	col_ind=col_ind_name[:,0:2].astype(int)
	col_name=col_ind_name[:,2]

	pick_col_ind_name=np.asarray([
		[0,4,'station'],
		[6,7,'p_first_motion'],
		[7,8,'p_weight_code'],
		[58,62,'sr_dist_km'],
		[62,66,'takeoff'],
		[75,78,'azimuth'],
		[79,82,'takeoff_uncertainty'],
		[83,86,'azimuth_uncertainty'],
		[95,98,'channel'],
	])
	pick_col_ind=pick_col_ind_name[:,0:2].astype(int)
	pick_col_name=pick_col_ind_name[:,2]

	header_all=[]
	pick_all=[]
	with open(fpfile, 'r') as file1:
		is_header_line=True
		for line in file1:
			if is_header_line:
				header_all.append([line[i:j] for i,j in zip(col_ind[:,0], col_ind[:,1])])
				header_all[-1][-1]=header_all[-1][-1].strip()
				event_id=header_all[-1][-1]
				tmp_pol=[]
				is_header_line=False
			else:
				if line[0:3]=='   ': # Footer line
					tmp_pick_df=pd.DataFrame(tmp_pol,columns=pick_col_name)

					# Drops any phase arrivals that do not have a P first motion
					tmp_pick_df=tmp_pick_df.drop(tmp_pick_df.loc[tmp_pick_df['p_first_motion']==' '].index)
					# Adds event_id to the dataframe
					tmp_pick_df['event_id']=str(event_id)

					pick_all.append(tmp_pick_df)
					is_header_line=True
				else: # Phase arrival line
					tmp_pol.append([line[i:j] for i,j in zip(pick_col_ind[:,0], pick_col_ind[:,1])])

	cat_df=pd.DataFrame(header_all,columns=col_name)
	cat_df=cat_df.astype({  'year':'int32','month':'int32',
							'day':'int32','second':'float',
							'lat_deg':'int32','lat_min':'int32',
							'lon_deg':'int32','lon_min':'int32',
							'depth':'int32','x_error_km':'int32',
							'z_error_km':'int32','mag_pref':'int32'})

	cat_df.loc[cat_df['year']>50,'year']+=1900
	cat_df.loc[cat_df['year']<50,'year']+=2000
	cat_df.loc[cat_df['hour']=='  ','hour']=0
	cat_df.loc[cat_df['minute']=='  ','minute']=0
	cat_df['hour']=cat_df['hour'].astype('int32')
	cat_df['minute']=cat_df['minute'].astype('int32')
	cat_df['second']/=100
	cat_df['origin_DateTime']=pd.to_datetime(cat_df.loc[:,['year','month','day','hour','minute','second']])
	cat_df['origin_lat']=cat_df['lat_deg']+cat_df['lat_min']/6000
	cat_df.loc[cat_df['latNS']=='S','origin_lat']*=-1
	cat_df['origin_lon']=cat_df['lon_deg']+cat_df['lon_min']/6000
	cat_df.loc[cat_df['lonEW']!='E','origin_lon']*=-1
	cat_df['origin_depth_km']=cat_df['depth']/100
	cat_df['horz_uncert_km']=cat_df['x_error_km']/100
	cat_df['vert_uncert_km']=cat_df['z_error_km']/100
	cat_df['event_mag']=cat_df['mag_pref']/10

	cat_df=cat_df.loc[:,['origin_DateTime', 'origin_lat', 'origin_lon', 'origin_depth_km',
						'horz_uncert_km', 'vert_uncert_km', 'event_mag', 'event_id']]

	pick_df=pd.concat(pick_all).reset_index(drop=True)
	pick_df['takeoff_uncertainty']=pick_df['takeoff_uncertainty'].replace('   ',0)
	pick_df['azimuth_uncertainty']=pick_df['azimuth_uncertainty'].replace('   ',0)
	pick_df=pick_df.astype({  'p_weight_code':'int32','sr_dist_km':'float','takeoff':'int32',
							'azimuth':'int32','takeoff_uncertainty':'int32','azimuth_uncertainty':'int32'})
	pick_df['p_polarity']=0.
	pick_df['sr_dist_km']/=10
	pick_df['p_first_motion']=pick_df['p_first_motion'].str.upper()

	pick_df['takeoff']=180-pick_df['takeoff']

	pick_df.loc[pick_df['p_weight_code']==0,'p_polarity']=1.
	pick_df.loc[pick_df['p_weight_code']==1,'p_polarity']=0.5
	pick_df.loc[pick_df['p_weight_code']==2,'p_polarity']=0.2
	pick_df.loc[pick_df['p_weight_code']==3,'p_polarity']=0.1
	pick_df.loc[(pick_df['p_first_motion']=='D') | (pick_df['p_first_motion']=='-'),'p_polarity']*=-1

	pick_df=pick_df.drop(columns=['p_first_motion','p_weight_code'])

	pick_df['station']=pick_df['station'].str.strip()
	pick_df['channel']=pick_df['channel'].str.strip()

	return cat_df,pick_df


def read_hash3_polarity_file(fpfile,p_weight_I=1.0,p_weight_E=0.5,sta_name_length=4):
	'''
	Reads input phase file following the old SCEC format.
	Gives Impulsive (I) manual picks a weight of p_weight_I (default 1.0)
	Gives Emmergent (E) manual picks a weight of p_weight_E (default 0.5)
	sta_name_length: How many characters the station name has been allocated in the file.
	'''

	col_len=[[0,sta_name_length],
			[sta_name_length+1,sta_name_length+3],
			[sta_name_length+5,sta_name_length+8],
			[sta_name_length+9,sta_name_length+10],
			[sta_name_length+11,sta_name_length+12],
			]

	file1 = open(fpfile, 'r')
	header_list=[]
	catalog_polarity_list=[]
	event_polarity_list=[]
	is_header=True
	for line in file1:
		if is_header:
			origin_year=int(line[0:4])
			origin_month=int(line[4:6])
			origin_day=int(line[6:8])
			origin_hour=int(line[8:10])
			origin_minute=int(line[10:12])
			origin_second=float(line[12:17])

			origin_lat_deg=int(line[17:19])
			origin_lat_NS=line[19]
			origin_lat_min=float(line[20:25])
			origin_lon_deg=int(line[25:28])
			origin_lon_EW=line[28]
			origin_lon_min=float(line[29:34])

			origin_depth_km=float(line[34:39])
			horz_uncert_km=float(line[88:93])
			vert_uncert_km=float(line[94:99])
			event_mag=float(line[139:143])
			mag_type='M?'
			event_id=str(line[149:165]).strip()

			if origin_lat_NS!='S':
				origin_lat_NS='N'
			if origin_lon_EW!='E':
				origin_lon_EW='W'
			origin_lat=dm2dd(origin_lat_deg,origin_lat_min,origin_lat_NS)
			origin_lon=dm2dd(origin_lon_deg,origin_lon_min,origin_lon_EW)
			header_df=pd.DataFrame( data=[[origin_year,origin_month,origin_day,
											origin_hour,origin_minute,origin_second,
											origin_lat,origin_lon,origin_depth_km,
											horz_uncert_km,vert_uncert_km,
											event_mag,mag_type,event_id]],
									columns=['year','month','day',
											'hour','minute','second',
											'origin_lat','origin_lon','origin_depth_km',
											'horz_uncert_km','vert_uncert_km',
											'event_mag','mag_type','event_id'])
			header_list.append(header_df)
			is_header=False
		elif line[0:4]=='    ':
			tmp_df=pd.DataFrame(data=event_polarity_list,columns=['station','network','location','channel','p_polarity','event_id'])
			try:
				tmp_df=pd.DataFrame(data=event_polarity_list,columns=['station','network','location','channel','p_polarity','event_id'])
			except:
				raise ValueError('*** Error reading line:\n{}'.format(line))
			tmp_df['network']=tmp_df['network'].str.strip()
			tmp_df['station']=tmp_df['station'].str.strip()
			tmp_df['location']=tmp_df['location'].str.strip()
			tmp_df['channel']=tmp_df['channel'].str.strip()
			# tmp_df=tmp_df.sort_values(by=['station','network','channel','location']).reset_index(drop=True)
			catalog_polarity_list.append(tmp_df)
			event_polarity_list=[]
			is_header=True
		else:

			station_name=line[col_len[0][0]:col_len[0][1]]
			network_name=line[col_len[1][0]:col_len[1][1]]
			channel_name=line[col_len[2][0]:col_len[2][1]]
			p_onset=line[col_len[3][0]:col_len[3][1]]
			p_polarity=line[col_len[4][0]:col_len[4][1]]
			location_name='--'
			if (p_onset=='I') | (p_onset=='i'):

				p_weight=p_weight_I
			elif (p_onset=='E') | (p_onset=='e'):
				p_weight=p_weight_E
			else:
				print('Unknown p_onset (',p_onset,'). Discarding polarity.')
				p_weight=0

			if (p_polarity=='U') | (p_polarity=='u') | (p_polarity=='+'):
				p_polarity=p_weight
			elif (p_polarity=='D') | (p_polarity=='d') | (p_polarity=='-'):
				p_polarity=-p_weight
			else:
				print('Unknown p_polarity (',p_polarity,'). Discarding polarity.')
				p_weight=0

			if p_polarity!=0:
				event_polarity_list.append([station_name,network_name,location_name,channel_name,p_polarity,event_id])
	file1.close()

	cat_df=pd.concat(header_list).reset_index(drop=True)
	origin_DateTime=pd.to_datetime(cat_df.loc[:,['year','month','day','hour','minute','second']])
	cat_df.insert(0,'origin_DateTime',origin_DateTime)
	cat_df=cat_df.drop(columns=['year','month','day','hour','minute','second','mag_type'])

	tmp_pol_df=pd.concat(catalog_polarity_list).reset_index(drop=True)
	return cat_df,tmp_pol_df


def read_hash4_polarity_file(fpfile,p_weight_I=1.0,p_weight_E=0.5):
	'''
	Reads a polarity file following the HASH driver 4 format.
	Gives Impulsive (I) manual picks a weight of p_weight_I (default 1.0)
	Gives Emmergent (E) manual picks a weight of p_weight_E (default 0.5)
	'''
	col_ind_name=np.asarray([
		[0,4,'year'],
		[4,6,'month'],
		[6,8,'day'],
		[8,10,'hour'],
		[10,12,'minute'],
		[12,16,'second'],
		[16,18,'lat_deg'],
		[18,19,'latNS'],
		[19,23,'lat_min'],
		[23,26,'lon_deg'],
		[26,27,'lonEW'],
		[27,31,'lon_min'],
		[31,36,'depth'],
		[130,146,'event_id'],
		[147,150,'mag_pref']
	])
	col_ind=col_ind_name[:,0:2].astype(int)
	col_name=col_ind_name[:,2]

	pick_col_ind_name=np.asarray([
		[0,4,'station'],
		[5,7,'network'],
		[9,12,'channel'],
		[13,14,'p_onset'],
		[15,16,'p_first_motion'],
		[16,17,'p_weight_code'],
	])
	pick_col_ind=pick_col_ind_name[:,0:2].astype(int)
	pick_col_name=pick_col_ind_name[:,2]

	header_all=[]
	pick_all=[]
	with open(fpfile, 'r') as file1:
		is_header_line=True
		for line in file1:
			if line[14].upper() in ['P','S','+','-',' ']: # Phase arrival line
				tmp_pol.append([line[i:j] for i,j in zip(pick_col_ind[:,0], pick_col_ind[:,1])])
			else:
				if len(header_all)>0:
					tmp_pick_df=pd.DataFrame(tmp_pol,columns=pick_col_name)
					# Drops any phase arrivals that do not have a P first motion
					tmp_pick_df=tmp_pick_df.drop(tmp_pick_df.loc[tmp_pick_df['p_first_motion']==' '].index)
					# Adds event_id to the dataframe
					tmp_pick_df['event_id']=str(event_id)
					pick_all.append(tmp_pick_df)

				header_all.append([line[i:j] for i,j in zip(col_ind[:,0], col_ind[:,1])])
				header_all[-1][-2]=header_all[-1][-2].strip()
				event_id=header_all[-1][-2]
				tmp_pol=[]
				is_header_line=False
		if len(tmp_pol)>0:
			tmp_pick_df=pd.DataFrame(tmp_pol,columns=pick_col_name)
			# Drops any phase arrivals that do not have a P first motion
			tmp_pick_df=tmp_pick_df.drop(tmp_pick_df.loc[tmp_pick_df['p_first_motion']==' '].index)
			# Adds event_id to the dataframe
			tmp_pick_df['event_id']=event_id
			pick_all.append(tmp_pick_df)

	cat_df=pd.DataFrame(header_all,columns=col_name)
	cat_df=cat_df.astype({  'year':'int32','month':'int32',
							'day':'int32','second':'float',
							'lat_deg':'int32','lat_min':'int32',
							'lon_deg':'int32','lon_min':'int32',
							'depth':'int32','mag_pref':'int32'})

	cat_df['second']/=100
	cat_df['origin_DateTime']=pd.to_datetime(cat_df.loc[:,['year','month','day','hour','minute','second']])
	cat_df['origin_lat']=cat_df['lat_deg']+cat_df['lat_min']/6000
	cat_df.loc[cat_df['latNS']=='S','origin_lat']*=-1
	cat_df['origin_lon']=cat_df['lon_deg']+cat_df['lon_min']/6000
	cat_df.loc[cat_df['lonEW']!='E','origin_lon']*=-1
	cat_df['origin_depth_km']=cat_df['depth']/100
	cat_df['event_mag']=cat_df['mag_pref']/100

	cat_df['horz_uncert_km']=0
	cat_df['vert_uncert_km']=0
	cat_df=cat_df.loc[:,['origin_DateTime', 'origin_lat', 'origin_lon', 'origin_depth_km',
						'horz_uncert_km', 'vert_uncert_km', 'event_mag', 'event_id']]

	pick_df=pd.concat(pick_all).reset_index(drop=True)
	pick_df['p_onset']=pick_df['p_onset'].str.upper()
	pick_df['p_first_motion']=pick_df['p_first_motion'].str.upper()
	pick_df['p_first_motion']=pick_df['p_first_motion'].replace('+','U')
	pick_df['p_first_motion']=pick_df['p_first_motion'].replace('-','D')

	pick_df['p_polarity']=0.
	pick_df.loc[pick_df['p_onset']=='I','p_polarity']=p_weight_I
	pick_df.loc[pick_df['p_onset']=='E','p_polarity']=p_weight_E
	pick_df.loc[(pick_df['p_first_motion']=='D'),'p_polarity']*=-1

	pick_df['station']=pick_df['station'].str.strip()
	pick_df['network']=pick_df['network'].str.strip()
	pick_df['channel']=pick_df['channel'].str.strip()

	pick_df=pick_df.drop(columns=['p_first_motion','p_onset'])

	return cat_df,pick_df

def get_element_value(element,name,dtype):
	'''
	Gets the value from a given element with the desired dtype.
	Used for parsing quakeml files.
	'''
	el = x_path(name,element)
	if len(el):
		el = x_path(name,element)[0]
		value = xpath2obj('value', el, dtype)
	else:
		value=np.nan

	return value

def xpath2obj(xpath, element=None, convert_to=str):
	'''
	Modified function for handling quakemls from Obspy:
	https://docs.obspy.org/_modules/obspy/io/quakeml/core.htm
	'''
	q = x_path(xpath, element=element)
	if not q:
		return None
	text = q[0].text
	if text is None or text == '':
		return None
	try:
		return convert_to(text)
	except Exception:
		print("*WARNING: Could not convert {} to type {}. Returning None.".format(text,convert_to))
	return None

def x_path(xpath, xml_root=None, element=None, namespace=None):
	'''
	Modified function for handling quakemls from Obspy:
	https://docs.obspy.org/_modules/obspy/io/quakeml/core.htm
	'''
	if element is None:
		element = xml_root

	namespaces = None
	if namespace:
		xpath = "b:%s" % xpath
		namespaces = {"b": namespace}
	elif hasattr(element, "nsmap") and None in element.nsmap:
		xpath = "b:%s" % xpath
		namespaces = {"b": element.nsmap[None]}
	return element.xpath(xpath, namespaces=namespaces)

def get_value(values):
	if len(values)>0:
		value=values[0]
	else:
		value=None
	return value

def read_quakeml_polarity_file(fpfile,p_weight_I=1.0,p_weight_E=0.5):
	'''
	Reads catalog and polarity information from a quakeml polarity file.
	Precomputed takeoff and source-receiver azimuths are expected.

	Gives Impulsive picks a weight of p_weight_I (default 1.0)
	Gives Emmergent picks a weight of p_weight_E (default 0.5)
	'''
	try:
		from lxml import etree
	except:
		raise ImportError('Cannot import lxml, a requirement for handling quakeml files. Did you install this package?')

	xml_root=etree.parse(fpfile).getroot()
	namespace=etree.QName(xml_root[0].tag).namespace
	catalog_el=x_path('eventParameters', xml_root=xml_root,namespace=namespace)[0]

	cat_df=pd.DataFrame(data={'event_id':np.array([],dtype=str),'origin_DateTime':np.array([],dtype='datetime64[ns]'),
								'event_mag':np.array([],dtype=float),'origin_lat':np.array([],dtype=float),
								'origin_lon':np.array([],dtype=float),'origin_depth_km':np.array([],dtype=float)})

	pick_list=[]
	for event_x,event_el in enumerate(x_path('event', catalog_el)):
		# Gets event_id
		event_resource_id = event_el.get('publicID')
		event_id=''.join(event_resource_id.split('/')[-2:]).lower()

		# Gets origin element
		origin_el=get_value(x_path('origin', event_el))

		time=get_element_value(origin_el,'time',np.datetime64)
		latitude=get_element_value(origin_el,'latitude',float)
		longitude=get_element_value(origin_el,'longitude',float)
		depth=get_element_value(origin_el,'depth',float)

		# Gets magnitude element
		magnitude_el=get_value(x_path('magnitude', event_el))
		magnitude=-999.
		if not(type(magnitude_el)==type(None)):
			if len(magnitude_el):
				magnitude=get_element_value(magnitude_el,'mag',float)

		# Records origin and magnitude info in the catalog dataframe
		cat_df.loc[event_x,'event_id']=event_id
		cat_df.loc[event_x,'origin_DateTime']=time
		cat_df.loc[event_x,'event_mag']=magnitude
		cat_df.loc[event_x,'origin_lat']=latitude
		cat_df.loc[event_x,'origin_lon']=longitude
		cat_df.loc[event_x,'origin_depth_km']=depth/1000

		# Gets pick information from the event element and records it in a dataframe
		pick_df=pd.DataFrame(data={'network':np.array([],dtype=str),'station':np.array([],dtype=str),
									'location':np.array([],dtype=str),'channel':np.array([],dtype=str),
									'phase':np.array([],dtype=str),
									'polarity':np.array([],dtype=str),'onset':np.array([],dtype=str),
									'takeoff':np.array([],dtype=float),'azimuth':np.array([],dtype=float)})

		for pick_x,pick_el in enumerate(x_path('pick', event_el)):
			for wid_el in x_path('waveformID', pick_el):
				pick_df.loc[pick_x,'network']=wid_el.get('networkCode').strip() or ''
				pick_df.loc[pick_x,'station']=wid_el.get('stationCode').strip() or ''
				pick_df.loc[pick_x,'location']=wid_el.get('locationCode').strip() or ''
				pick_df.loc[pick_x,'channel']=wid_el.get('channelCode').strip() or ''

			pick_df.loc[pick_x,'phase']=xpath2obj('phaseHint',pick_el) or xpath2obj('phase',pick_el)
			pick_df.loc[pick_x,'polarity']=xpath2obj('polarity',pick_el)
			pick_df.loc[pick_x,'onset']=xpath2obj('onset',pick_el)

		for arrival_x,arrival_el in enumerate(x_path('arrival', origin_el)):
			pick_df.loc[arrival_x,'azimuth']=xpath2obj('azimuth', arrival_el, float)
			pick_df.loc[arrival_x,'takeoff']=get_element_value(arrival_el,'takeoffAngle',float)
			if np.isnan(pick_df.loc[arrival_x,'takeoff']):
				pick_df.loc[arrival_x,'takeoff']=xpath2obj('takeoffAngle', arrival_el, float)
			pick_df.loc[arrival_x,'sr_dist_km']=xpath2obj('distance', arrival_el, float)
			pick_df.loc[arrival_x,'phase']=xpath2obj('phase',arrival_el,str)

		pick_df['event_id']=event_id
		pick_list.append(pick_df)
	pick_df=pd.concat(pick_list).reset_index(drop=True)

	# Selects P picks
	pick_df=pick_df[(pick_df['phase']=='P')]

	# Considers any polarity string other than 'positive' or 'negative' to be null
	pick_df.loc[~pick_df['polarity'].isin(['positive','negative']),'polarity']=None

	# If a polarity has missing takeoff/azimuth/sr_dist_km, looks for that info elsewhere
	if (~pd.isnull(pick_df['polarity']) & pd.isnull(pick_df[['takeoff','azimuth','sr_dist_km']]).any(axis=1)).any():
		# Selects picks with takeoff/azimuth/sr_dist_km info
		ta_pick_df=pick_df[(~pd.isnull(pick_df[['azimuth','takeoff','sr_dist_km']])).all(axis=1)]
		ta_pick_df=ta_pick_df.drop_duplicates(subset=['network','station','location','channel','event_id'])

		# Selects picks with polarity info
		pol_pick_df=pick_df.loc[~pd.isnull(pick_df['polarity'])].copy()

		pol_pick_df[['azimuth','takeoff','sr_dist_km']]=(pol_pick_df.merge(ta_pick_df[['network','station','location','channel','azimuth','takeoff','sr_dist_km','event_id']],on=['network','station','location','channel','event_id'],how='left'))[['azimuth_y','takeoff_y','sr_dist_km_y']].values

		pick_df.loc[pol_pick_df.index,['azimuth','takeoff','sr_dist_km']]=pol_pick_df[['azimuth','takeoff','sr_dist_km']].values

	# # Considers only impulsive or emergent arrivals
	# pick_df=pick_df.loc[pick_df['onset'].isin(['impulsive','emergent']),:].reset_index(drop=True)

	# Considers only arrivals that have polarity information
	pick_df=pick_df.drop(pick_df[pd.isnull(pick_df['polarity'])].index).reset_index(drop=True)

	# Considers only arrivals that have reported takeoff and azimuth information
	ta_null_index=pick_df[pd.isnull(pick_df[['takeoff','azimuth']]).any(axis=1)].index
	if len(ta_null_index):
		print('*WARNING: {} arrival(s) in the quakeml file ({}) are missing takeoff/azimuthal information. Ignoring these arrivals.'.format(len(ta_null_index),fpfile))
		pick_df=pick_df.drop(ta_null_index).reset_index(drop=True)

	# Drops duplicate records
	pick_df=pick_df.drop_duplicates().reset_index(drop=True)

	pick_df['p_polarity']=0.
	pick_df.loc[pick_df['polarity']=='positive','p_polarity']=1.
	pick_df.loc[pick_df['polarity']=='negative','p_polarity']=-1.

	# Drops any polarities with 0 weight
	pick_df=pick_df.drop(pick_df[pick_df['p_polarity']==0].index).reset_index(drop=True)

	if p_weight_I!=1:
		pick_df.loc[pick_df['onset']=='impulsive','p_polarity']*=p_weight_I
	if p_weight_E!=1:
		pick_df.loc[pick_df['onset']=='emergent','p_polarity']*=p_weight_E

	# Drops unneeded columns
	pick_df=pick_df.drop(columns=['polarity','onset'])

	return cat_df,pick_df